from setuptools import setup, find_packages
from numpy.distutils.core import setup, Extension


ext1 = Extension(
    name="gnssrefl.gpssnr",
    sources=["gnssrefl/gpssnr.f"],
    f2py_options=["--verbose"],
)

ext2 = Extension(
    name="gnssrefl.gnsssnr",
    sources=["gnssrefl/gnsssnr.f"],
    f2py_options=["--verbose"],
)
ext3 = Extension(
    name="gnssrefl.gnsssnrbigger",
    sources=["gnssrefl/gnsssnrbigger.f"],
    f2py_options=["--verbose"],
)
ext4 = Extension(
    name="gnssrefl.xnmeasnr",
    sources=["gnssrefl/xnmeasnr.f"],
    f2py_options=["--verbose"],
)

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = [
    "numpy",
    "wget",
    "scipy",
    "matplotlib",
    "requests",
    "progress",
    "astropy",
    "simplekml",
    "earthscope-sdk",
    "jupyterlab",
    "ipywidgets"
]
setup(
    name="gnssrefl",
    version="3.6.3",
    author="Kristine Larson",
    author_email="kristinem.larson@gmail.com",
    description="A GNSS reflectometry software package ",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/kristinemlarson/gnssrefl/",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "gnssir = gnssrefl.gnssir_cl:main",
            "gnssir_old = gnssrefl.gnssir_cl_old:main",
            "rinex2snr = gnssrefl.rinex2snr_cl:main",
            "daily_avg = gnssrefl.daily_avg_cl:main",
            "quickLook= gnssrefl.quickLook_cl:main",
            "download_rinex = gnssrefl.download_rinex:main",
            "download_orbits = gnssrefl.download_orbits:main",
            "gnssir_input = gnssrefl.gnssir_input:main",
            "ymd = gnssrefl.ymd:main",
            "ydoy = gnssrefl.ydoy:main",
            "xyz2llh = gnssrefl.xyz2llh:main",
            "llh2xyz = gnssrefl.llh2xyz:main",
            "prn2gps = gnssrefl.prn2gps:main",
            "download_tides = gnssrefl.download_tides:main",
            "subdaily= gnssrefl.subdaily_cl:main",
            "gpsweek = gnssrefl.gpsweek:main",
            "nmea2snr= gnssrefl.nmea2snr_cl:main",
            "installexe= gnssrefl.installexe_cl:main",
            "download_unr = gnssrefl.download_unr:main",
            "query_unr= gnssrefl.query_unr:main",
            "mp1mp2= gnssrefl.computemp1mp2:main",
            "download_teqc = gnssrefl.download_teqc:main",
            "rinex3_rinex2= gnssrefl.rinex3_rinex2:main",
            "veg_multiyr= gnssrefl.veg_multiyr:main",
            "check_rinex_file= gnssrefl.check_rinex_file:main",
            "rinex3_snr= gnssrefl.rinex3_snr:main",
            "rt_rinex3_snr= gnssrefl.rt_rinex3_snr:main",
            "filesizes= gnssrefl.filesizes:main",
            "invsnr= gnssrefl.invsnr_cl:main",
            "invsnr_input= gnssrefl.invsnr_input:main",
            "vwc_input= gnssrefl.vwc_input:main",
            "phase= gnssrefl.quickPhase:main",
            "refl_zones= gnssrefl.refl_zones_cl:main",
            "rinex_coords= gnssrefl.rinex_coords:main",
            "vwc= gnssrefl.vwc_cl:main",
            "smoosh= gnssrefl.smoosh:main",
            "smoosh_snr= gnssrefl.smoosh_snr:main",
            "quickplt= gnssrefl.quickplt:main",
            "snowdepth= gnssrefl.snowdepth_cl:main",
            "rh_plot= gnssrefl.rh_plot:main",
            "max_resolve_RH= gnssrefl.max_resolve_RH_cl:main",
            "pickle_dilemma= gnssrefl.pickle_dilemma:main",
            "mjd= gnssrefl.mjd:main",
            "make_meta= gnssrefl.make_meta:main",
        ],
    },
    install_requires=requirements,
    ext_modules=[ext1, ext2, ext3, ext4],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



# -*- coding: utf-8 -*-
import argparse
import sys

import gnssrefl.gps as g

def main():
    """
    Converts Cartesian coordinates to latitude, longitude, ellipsoidal height.
    Prints to screen

    Example
    -------
    xyz2llh -1283634.1615 -4726427.8931 4074798.0429
        returns 39.949492042 -105.194266387  1728.856

    Parameters
    ----------
    x : float
        X coordinate (m)
    y : float
        Y coordinate (m)
    z : float
        Z coordinate (m)

    """

    parser = argparse.ArgumentParser()
    parser.add_argument("x", help="X coordinate (m) ", type=float)
    parser.add_argument("y", help="Y coordinate (m) ", type=float)
    parser.add_argument("z", help="Z coordinate (m) ", type=float)

    args = parser.parse_args()

    x=args.x; y=args.y; z=args.z
    xyz = [x, y, z]
    lat,lon,h = g.xyz2llhd(xyz)
    print("Lat Lon Ht (deg deg m) %14.9f %14.9f %9.3f " % (lat, lon, h) )

if __name__ == "__main__":
    main()



import argparse
import datetime
import matplotlib.pyplot as plt
import numpy as np
import os
import sys

from datetime import date

import gnssrefl.gps as g

def main():
    """
    very simple code to pick up all the file sizes for SNR files in a given year
    only checks for snr66 files.  

    this is too slow - instead of using numpy array - use a list - and then change to
    numpy array at the end

    It makes a plot - not very useful now that files are nominal gzipped.

    Parameters
    ----------
    station : str
        4 character station name
    year1 : int, optional
        beginning year
    year2 : int, optional
        ending year
    doy1 : int, optional
        beginning day of year
    doy2 : int, optional
        ending day of year
    gz : bool, optional
        say T or True to search for gzipped files

    """
    g.check_environ_variables()
    xdir = os.environ['REFL_CODE'] 

# must input start and end year
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name", type=str)
    parser.add_argument("-year1", default=None, type=str, help="restrict to years starting with")
    parser.add_argument("-year2", default=None, type=str, help="restrict to years ending with")
    parser.add_argument("-doy1", default=None, type=str, help="restrict to after doy in year1")
    parser.add_argument("-doy2", default=None, type=str, help="restrict to before doy in year2")
    parser.add_argument("-gz", default=None, type=str, help="Boolean to search for gz files")
    args = parser.parse_args()
#   these are required
    station = args.station
    gz = args.gz

    if args.year1 == None:
        year1 = 2005
    else:
        year1=int(args.year1)

    if args.year2 == None:
        year2 = 2030
    else:
        year2=int(args.year2)

    if args.doy1== None:
        doy1=1
    else:
        doy1 = int(args.doy1)

    if args.doy2== None:
        doy2=366
    else:
        doy2 = int(args.doy2)

    tstart = year1+ doy1/365.25
    tend   = year2+ doy2/365.25

    gzadd = ''
    if (gz == 'T') or (gz == 'True'):
        gzadd = '.gz'

    k=0
# added standard deviation 2020 feb 14, changed n=6
# now require it as an input
# you can change this - trying out 80 for now
#ReqTracks = 80
# putting the results in a np.array, year, doy, RH, Nvalues, month, day
    n=3
    tv = np.empty(shape=[0, n])
    obstimes = []; 
    year_list = np.arange(year1, year2+1, 1)
    for yr in year_list:
        direc = xdir + '/' + str(yr) + '/snr/' + station + '/' 
        for doy in range(0,367):
            year, month, day, cyyyy,cdoy, YMD = g.ydoy2useful(yr,doy)

            fname = direc + station + cdoy + '0.' + cyyyy[2:4]  + '.snr66' + gzadd
            #print(fname)
            t = yr+doy/365.25
            if os.path.isfile(fname):
                a = np.loadtxt(fname,skiprows=3,comments='%')
                nr,nc=a.shape # nr is number of obs
                filler = datetime.datetime(year=yr, month=month, day=day)
                # this is for the daily average
                newl = [yr, doy, nr]
                if (t >= tstart) & (t <= tend):
                    #print(nr, year, doy)
                    tv = np.append(tv, [newl],axis=0)
                    obstimes.append(filler)
            else:
                if (t >= tstart) & (t <= tend):
                    newl = [yr, doy, 0]
                    tv = np.append(tv, [newl],axis=0)
                    filler = datetime.datetime(year=yr, month=month, day=day)
                    obstimes.append(filler)

    fs = 12
    fig,ax=plt.subplots()
    plt.plot(obstimes, tv[:,2]/1000,'b.')
    fig.autofmt_xdate()
    plt.ylabel('nobs/1000',fontsize=fs)
    plt.title('Observations for Station: ' + station,fontsize=fs)
    plt.yticks(fontsize=fs)
    plt.xticks(fontsize=fs)
    plt.grid()
    plt.show()


if __name__ == "__main__":
    main()



#
# Author: Kristine M. Larson
# Date: december 25, 2021
# Purpose: help set up json input file needed for snr inversion code

import argparse
import json
import os
import subprocess
import sys

# internal library of GPS functions
import gnssrefl.gps as g

# function to read UNR database
# import unr


def parse_arguments():
# user inputs the observation file information
    parser = argparse.ArgumentParser()
# required arguments
    parser.add_argument("station", help="station name", type=str)
    parser.add_argument("-h1", default=None, type=float, help="Lower limit reflector height (m), default 1")
    parser.add_argument("-h2", default=None, type=float, help="Upper limit reflector height (m), default 8")
    parser.add_argument("-e1", default=None, type=float, help="Lower limit elev. angle (deg), default 5")
    parser.add_argument("-e2", default=None, type=float, help="Upper limit elev. angle (deg), default 15")

    parser.add_argument("-azim1", default=None, type=float, help="Lower limit azimuth angle (deg), default 0")
    parser.add_argument("-azim2", default=None, type=float, help="Upper limit azimuth angle (deg), default 360")
    parser.add_argument("-peak2noise", default=None, type=float, help="Peak2noise QC criterion")

    parser.add_argument("-lat", help="Latitude (degrees)", type=str, default=None)
    parser.add_argument("-lon", help="Longitude (degrees)", type=str, default=None)
    parser.add_argument("-height", help="Ellipsoidal height (meters)", type=str, default=None)
# these are the optional inputs 
    args = parser.parse_args().__dict__

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def invsnr_input(station: str, h1: float=1, h2: float=8, e1: float=5, e2: float=15, azim1: float = 0,
                 azim2: float = 360, lat: float = None, lon: float = None, height: float = None, peak2noise: float=3):
    """
    Sets some of the analysis parameters for invnsr. Values are stored in a json in $REFL_CODE/input
    Note: this code was written independently of gnssrefl. The Quality Control parametesr are thus quite 
    different from how gnssrefl is done. The LSP RH retrievals are not - and should not - be the same.

    Examples
    --------
    invsnr_input sc02 -h1 3 -h2 12 -e1 5 -e2 13 -azim1 40 -azim2 220
        general Friday Harbor inputs
    invsnr_input at01 -h1 9 -h2 14 -e1 5 -e2 13 -azim1 20 -azim2 22
        St Michael inputs 



    Parameters
    ----------
    station : str
        four ch ID of the station
    h1 : float, optional
        Lower limit reflector height (m)
    h2 : float, optional
        Upper limit reflector height (m)
    e1 : float, optional
        Lower limit elev. angle (deg)
    e2 : float
        Upper limit elev. angle (deg)
    azim1 : float 
        Lower limit azimuth angle (deg)
    azim2 : float 
        Upper limit azimuth angle (deg)
    lat : float, optional
        Latitude (degrees)
    lon : float, optional
        Longitude (degrees)
    height : float, optional
        Ellipsoidal height (meters)
    peak2noise : float, optional
        peak to noise 

    """

# rename the user inputs into variables
#
    # use old variable names
    a1=azim1
    a2=azim2
    NS = len(station)
    if (NS != 4):
        print('station name must be four characters long. Exiting.')
        sys.exit()
    Lat, Long, Height = g.queryUNR_modern(station)
    if Lat == 0:
        if lat is None:
            print('I did not find your station at UNR.')
            print('You need to set them using -lat, -lon, -height')
            sys.exit()
        else:
            Lat = float(lat)
            Long = float(lon)
            Height = float(height)

# start the lsp dictionary (left over name from another piece of code)
    lsp={}
    lsp['station'] = station
    lsp['lat'] = Lat; lsp['lon'] = Long; lsp['ht']=Height
#
    if h1 > h2:
        print('h1 cannot be greater than h2. ', h1, h2, ' Exiting.')
        sys.exit()

    if e1 > e2:
        print('e1 must be less than e2.', e1, e2, ' Exiting.')
        sys.exit()

    if a1 > a2:
        print('a1 must be less than a2.', a1, a2, ' Exiting.')
        sys.exit()

    # make groupings of RH, ele, and azim limits
    lsp['rhlims'] = [h1, h2]
    lsp['elvlims'] = [e1, e2]
    lsp['azilims'] = [a1, a2]
    # use david's naming convention for this
    lsp['pktnlim'] = peak2noise

# where the instructions will be written eventually. for now in the current directory
    if 'REFL_CODE' not in os.environ:
        print('The REFL_CODE environment variable has not been set, Please take care of this.')
        print('Exiting')
        sys.exit()
    xdir = os.environ['REFL_CODE']
    outputdir  = xdir + '/input' 
    if not os.path.isdir(outputdir):
        subprocess.call(['mkdir', outputdir])
# 
    lsp['precision'] = 0.005 # precision of RH in meters
# 
    lsp['l2c_only'] = True

    outputfile = outputdir + '/' + station + '.inv.json'
    print('Writing json file to:', outputfile)
    #print(lsp)
    with open(outputfile, 'w+') as outfile:
        json.dump(lsp, outfile, indent=4)


def main():
    args = parse_arguments()
    invsnr_input(**args)

if __name__ == "__main__":
    main()



#reflection zones command line function

import argparse
import numpy as np
import os
import subprocess
import sys

import gnssrefl.gps as g
import gnssrefl.refl_zones as rf


def parse_arguments():
    parser = argparse.ArgumentParser()

    parser.add_argument('station', help='station', type=str)
    parser.add_argument('-lat', help='Latitude (deg), if station not in database', type=float, default=None)
    parser.add_argument('-lon', help='Longitude (deg), if station not in database', type=float, default=None)
    parser.add_argument('-height', help='Ellipsoidal height (m) if station not in database', type=float,default=None)
    parser.add_argument('-fr', help='1, 2, or 5 (default is 1)', type=int,default=None)
    parser.add_argument('-RH', help='Reflector height (meters). Default is height above sea level', type=float, default=None)
    parser.add_argument('-azim1', help='start azimuth (default is 0 deg, negative values allowed) ', type=int,default=None)
    parser.add_argument('-azim2', help='end azimuth (default is 360 deg) ', type=int,default=None)
    parser.add_argument('-el_list', nargs="*",type=float,  help='elevation angle list, e.g. 5 10 15  (default)')
    parser.add_argument('-azlist', nargs="*",type=float,  help='flexible azimuth angle list, e.g. 0 90 270 360, but must be positive ')
    parser.add_argument('-system', help='default is gps, options are galileo, glonass, beidou', type=str)
    parser.add_argument('-output', help='output filename. default is the station name with kml extension', type=str,default=None)

    args = parser.parse_args().__dict__


    return {key: value for key, value in args.items() if value is not None}

def reflzones(station: str, azim1: int=0, azim2: int=360, lat: float=None, lon: float=None, height: float=None, 
        RH: float=None, fr: int = 1, el_list: float= [], azlist : float=[], system: str = 'gps', output: str = None):
    """
    This module creates “stand-alone” Fresnel Zones maps for google Earth. At a minimum 
    it requires a four station character name as input. The output is a KML file.

    If the station is in the UNR database, those latitude, longitude, and ellipsoidal 
    height values are used. You may override those values with the optional inputs.

    The output file will be stored in REFL_CODE/Files/kml unless you specify an output name. In that case
    it will go into your working directory

    The defaults are that it does all azimuths, elevation angles of 5-10-15, GPS L1, and uses the height
    of the station above sea level to use for the RH. If you want to specify the reflector height, set -RH. 
    If you are making a file for an interior lake or river, you will need to use this option. Similarly, 
    for a soil moisture or snow reflection zone map, where height above sea level is not important, 
    you will want to set the RH value accordingly.


    Examples
    --------
    refl_zones p041 -RH 2
        standard Fresnel zones for all azimuths, GPS and L1 frequency, RH of 2 meters

    refl_zones sc02 -fr 2 -azlist 40 240
        Fresnel zones for limited azimuths, GPS and L2 frequency. RH is mean sea level

    refl_zones p041 -RH 5 -system galileo
        Fresnel zones for reflector height of 5 meters and Galileo L1

    refl_zones xxxx -RH 5 -lat 40 -lon 120 -height 10
        Using a station not in the UNR database, so station position given
        Note that this is the ellipsoidal height.

    Parameters
    ----------
    station : str
        station name
    azim1 : int, optional
        min azimuth angle in deg
    azim2 : int, optional
        max azimuth angle in deg
    lat : float, optional
        latitude in deg
    lon : float, optional
        longitude in deg
    height : float, optional
        ellipsoidal height in m
    RH : float, optional
        user-defined reflector height (m)
        default is to use sea level as the RH
    fr : int, optional
        frequency (only 1,2, or 5 allowed)
    el_list : list of floats, optional 
        elevation angles desired (deg)
        default is 5, 10, 15
    azlist : list of floats, optional
        azimuth angle regions (deg) Must be in pairs, i.e. 0 90 180 270
    system : str, optional
        name of constellation (gps,glonass,galileo, beidou allowed)
        default is gps
    output : str, optional
        name for kml file 

    Returns
    -------
    Creates a KML file for Google Earth

    """

    el_height = height # changed name to be consistent with gnssir_input

    # changed name to be consistent with other codes
    az_sectors = azlist
    # check that you have the files for the orbits on your local system
    foundfiles = rf.save_reflzone_orbits()    
    if not foundfiles:
        print('The orbit files needed for this code were either not found')
        print('or not downloaded successfully. They should be in the REFL_CODE/Files directory') 
        sys.exit()


    # check that EGM96 file is in your local directory
    foundfile = g.checkEGM()
    if (RH is None) and (not foundfile):
        print('EGM96 file has not been found. It should be in the REFL_CODE/Files directory')

    #print(lat,lon,el_height)
    if (lat is None) & (lon is None):
    # check the station coordinates in our database from the station name
        lat, lon, el_height = g.queryUNR_modern(station)
        if (lat == 0) and (lon == 0):
            print('Exiting.')
            sys.exit()
        else:
            print('Using inputs:', lat, lon, el_height)
    else:
        print('Using inputs:', lat, lon, el_height)

    geoidC = g.geoidCorrection(lat,lon)
    # calculate height above sea level
    sealevel = el_height - geoidC
    
    if fr not in [1,2,5]:
        print('Illegal frequency chosen: ',fr)
        return
    else:
        if (system == 'galileo') & (fr == 2):
            print('Galileo does not have a L2 frequency.')
            sys.exit()
        elif (system == 'glonass') & (fr == 5):
            print('Glonass does not have a L5 frequency.')
            sys.exit()
        else:
            print('Using frequency: ', fr)

    # set the reflector height
    # default is to use sea level as reflector height.  if RH is set, use that instead
    if RH is None:
        h = sealevel
    else:
        h = RH

    if h < 0:
        print('This is an illegal RH: ',h, ' Exiting.')
        sys.exit()

    if h > 300:
        print('This is a very large reflector height value:', round(h,2), '(m). Try setting RH. Exiting.')
        sys.exit()

    print('Reflector height (m) ', np.round(h,3))

    # set the default
    if len(el_list) == 0:
        el_list = [5, 10, 15]
    
    if len(el_list) > 5:
        el_list = el_list[0:5]
        print('Elevation angle list is very long - reducing to five.')

    emax = 61 #max allowed elevation angle  

    #print(all(x < 50 for x in el_list))

    if not (all(x < emax for x in el_list)):
        print('emax must be lower than 60 degrees. Resubmit your request.')
        sys.exit()


    obsfile, it  = rf.set_system(system)
    print('The code should use this orbit file: ', obsfile )
    x,y,z=g.llh2xyz(lat,lon,el_height)
    recv=np.array([x,y,z])
    # calculate rising and setting arcs for this site and the requested constellation
    azlisttmp= rf.rising_setting_new(recv,el_list,obsfile)
    if len(az_sectors) == 0:
        print('Using one set of azimuths') 
        azlist = rf.set_final_azlist(azim1,azim2,azlisttmp)
    else:
        NL = len(az_sectors)
        if (np.mod(NL,2) != 0):
            print('Your azimuth sector list does not have pairs of azimuths. Exiting')
            sys.exit()
        for ii in range(0,NL):
            if (az_sectors[ii] < 0):
                print('Negative azimuths not allowed. Exiting')
                sys.exit()
            if (az_sectors[ii] > 360):
                print('Azimuths greater than 360 not allowed. Exiting')
                sys.exit()

        for ii in range(0,int(NL/2)):
            if (az_sectors[ii*2] > az_sectors[ii*2+1]):
                print('Sectors must have increasing azimuth values. Exiting')
                sys.exit()

        print('Using sectors')
        azlist = rf.set_azlist_multi_regions(az_sectors,azlisttmp)

    # figure out where the output will go
    # changed it so it  goes to a subdirectory called "kml" by default 


    # if you set a filename, the output will be written there in the user directory
    # you need to set the kml to the ending.
    # if you use defaults, it is written in 

    if output == None: 
    # first check that the Files output directory exists
        xdir = os.environ['REFL_CODE']
        outputdir = xdir + '/Files' 
        if not os.path.isdir(outputdir):
            subprocess.call(['mkdir',outputdir])
        outputdir = xdir + '/Files/kml'  
        if not os.path.isdir(outputdir):
            subprocess.call(['mkdir',outputdir])
        output = outputdir  + '/' + station  + '.kml'

    # make the KML map file
    # 2023jan19 add station location
    fz_maps = rf.make_FZ_kml(station,output,fr, el_list, h, lat,lon,azlist)

    if fz_maps == True:
        print('File is saved in ' , output)
    else:
        print('Something went wrong, and the kml file was not created')


def main():
    args = parse_arguments()
    print(args)
    data = reflzones(**args)


if __name__ == "__main__":
    main()




from astropy.time import Time
from astropy.timeseries import LombScargle
import datetime
import math
import matplotlib.pyplot as plt
import numpy as np
import os
import pickle
from scipy import interpolate
from scipy.optimize import least_squares
import scipy.signal as spectral

from matplotlib.dates import (date2num, DateFormatter)
import subprocess
import sys
import time

# my local functions
import gnssrefl.gps as g
import gnssrefl.refraction as refr


def make_wavelength_column(nr,snrdata,signal):
    """
    NEEDS DOCUMENTATION

    Parameters
    ----------
    nr : integer
        number of rows in snrdata

    snrdata : numpy array
        snrfile array

    signal : string
        frequency  'L1','L2, etc

    Returns
    --------
    onecolumn : one-d numpy array
        snr data for the requested signal
    """
    onecolumn = np.ones((nr,1))
    if signal == 'L2':
        onecolumn = 2*onecolumn
    if signal == 'L5':
        onecolumn = 5*onecolumn
    if signal == 'L6':
        onecolumn = 6*onecolumn
    if signal == 'L7':
        onecolumn = 7*onecolumn

    return onecolumn


def readklsnrtxt(snrfile, thedir, signal):
    """
    parses the contents of a snrfile.
    The file itself is read in a separate function now; 
    if SNR data are zero for a given signal, the row is eliminated

    As of Oct 28, 2023, gzip after reading SNR file

    Parameters
    ----------
    snrfile: str
        variable with the file contents
    thedir : str
        directory where it is located
    signal: str
        'L1', 'L2' etc.

    Returns
    -------
    snrdata : numpy array of floats . Columns defined as:
        0 : satellite, usual (100 added for glonass, 200 added for galileo)
        1 : elev angle, deg
        2 : azimuth angle, deg
        3 : time in seconds since GPS began
        4 : SNR data in db-Hz
        5 : new column with wavelength in it, in meters.  

    """
    # do a straight load of the file
    snrall = loadsnrfile(snrfile, thedir)
    subprocess.call(['gzip', thedir + snrfile])

    nr,nc = snrall.shape
    # this is for the frequency information

    # will add Beidou later
    # 22feb09 and now the day has come!

    if signal == 'L1':
        snrdata = snrall[:,[0,1,2,3,6]]
        onecolumn = make_wavelength_column(nr,snrdata,signal)
    elif signal == 'L2':
        snrdata = snrall[:,[0,1,2,3,7]]
        onecolumn = make_wavelength_column(nr,snrdata,signal)
    # make sure the "9th" column is there
    elif (signal == 'L5') and (nc > 8):
        snrdata = snrall[:,[0,1,2,3,8]]
        onecolumn = make_wavelength_column(nr,snrdata,signal)
    elif signal == 'L1+L2':
        l1 = snrall[:,[0,1,2,3,6]]
        l1col = make_wavelength_column(nr,snrall,'L1')
        l2 = snrall[:,[0,1,2,3,7]]
        l2col = make_wavelength_column(nr,snrall,'L2')
        #print('vertically stacking L1 and L2 data')
        snrdata = np.vstack((l1,l2))
        onecolumn = np.vstack((l1col,l2col))
    elif signal == 'L1+L5':
        l1 = snrall[:,[0,1,2,3,6]]
        l1col = make_wavelength_column(nr,snrall,'L1')
        l5 = snrall[:,[0,1,2,3,8]]
        l5col = make_wavelength_column(nr,snrall,'L5')
        #print('vertically stacking L1 and L5 data')
        snrdata = np.vstack((l1,l5))
        onecolumn = np.vstack((l1col,l5col))
    elif (signal == 'L1+L2+L5'):
        l1 = snrall[:,[0,1,2,3,6]]
        l1col = make_wavelength_column(nr,snrall,'L1')

        l2 = snrall[:,[0,1,2,3,7]]
        l2col = make_wavelength_column(nr,snrall,'L2')

        l5 = snrall[:,[0,1,2,3,8]]
        l5col = make_wavelength_column(nr,snrall,'L5')

        tmp = np.vstack((l1,l2))
        snrdata = np.vstack((tmp,l5))

        tmp = np.vstack((l1col,l2col))
        onecolumn = np.vstack((tmp,l5col))
    elif (signal == 'L1+L2+L5+L6+L7'):
        if (nc <=9):
            print('You requested L1+L2+L5+L6+L7')
            print('But you can only do L1+L2+L5 with this size SNR file. Exiting.')
            sys.exit()

        l1 = snrall[:,[0,1,2,3,6]]
        l1col = make_wavelength_column(nr,snrall,'L1')

        l2 = snrall[:,[0,1,2,3,7]]
        l2col = make_wavelength_column(nr,snrall,'L2')

        l5 = snrall[:,[0,1,2,3,8]]
        l5col = make_wavelength_column(nr,snrall,'L5')

        l6 = snrall[:,[0,1,2,3,5]]
        l6col = make_wavelength_column(nr,snrall,'L6')

        l7 = snrall[:,[0,1,2,3,9]]
        l7col = make_wavelength_column(nr,snrall,'L7')

        tmp1 = np.vstack((l1,l2)) # stack 1,2
        tmp2 = np.vstack((l5,l6)) # stack 5,6
        tmp3 = np.vstack((tmp1,tmp2)) #stack 1,2,5,6
        snrdata = np.vstack((tmp3,l7)) # and ... 

        tmp1 = np.vstack((l1col,l2col))
        tmp2 = np.vstack((l5col,l6col))
        tmp3 = np.vstack((tmp1,tmp2))
        # now put them together ...
        onecolumn = np.vstack((tmp3,l7col))
    elif signal == 'L6':
        snrdata = snrall[:,[0,1,2,3,5]]
        onecolumn = make_wavelength_column(nr,snrdata,signal)
    elif signal == 'L7':
        snrdata = snrall[:,[0,1,2,3,9]]
        onecolumn = make_wavelength_column(nr,snrdata,signal)
    else:
        print('Did not find what was requested or illegal frequency. Exiting.')
        sys.exit()


    # add frequency column 
    snrdata = np.hstack((snrdata, onecolumn))

    # remove the zeroes
    ii = (snrdata[:,4] != 0)
    #print('Num of obs before ', len(snrdata))
    snrdata = snrdata[ii,:]
    print('Num of non-zero obs ', len(snrdata))

    # this could be done in the loading function
    stryear = str(int(snrfile[9:11]) + 2000)
    strdoy = snrfile[4:7]
    stryday = stryear + ':' + strdoy + ':00:00:00'
    #tobj = Time(stryday, format='yday')
    #gbase = tobj.gps
    # put the time tags into fake GPS time (seconds since GPS began?)
    #snrdata[:, 3] = snrdata[:, 3] + gbase

    return snrdata

def glonasswlen(prn, signal):
    """
    Given PRN, returns glonass wavelength

    Parameters
    ----------
    prn : integer
        satellite number
    signal : string
         L1 or L2 for glonass

    Returns
    -------
    wavelength : float
        wavelength for the given signal

    """
    channel = [1, -4, 5, 6, 1, -4, 5, 6, -2, -7, 0, -1, -2, -7, 0, -1, 4, -3, 3, 2, 4, -3, 3, 2]
    offset = 101  # 101 onwards is glonass
    try:
        channel_t = np.array([channel[i-offset] for i in prn], dtype=float)
    except TypeError:
        channel_t = channel[prn-offset]
    if signal == 'L1':
        lcar = g.constants.c / (1602e06 + channel_t * 0.5625e06)
        #lcar = 299792458 / (1602e06 + channel_t * 0.5625e06)
    elif signal == 'L2':
        lcar = g.constants.c / (1246e06 + channel_t * 0.4375e06)
        #lcar = 299792458. / (1246e06 + channel_t * 0.4375e06)
    else:
        #print('signal not recognised')
        lcar = np.nan
    return lcar


def datetime2gps(dt):
    """
    Parameters
    ----------
    dt : datetime 

    Returns  
    -------
    gpstime : float

    """

    timeobj = Time(dt, format='datetime')
    gpstime = timeobj.gps
    return gpstime


def gps2datetime(gt):
    """
    needs documentation
    """
    timeobj = Time(gt, format='gps', scale='utc')
    dt = timeobj.datetime
    return dt


def gps2datenum(gt):
    """
    needs documentation

    Parameters
    ----------
    gt : float
        gps time 

    Returns
    -------
    dn : datetime?

    """

    timeobj = Time(gt, format='gps', scale='utc')
    dt = timeobj.datetime
    dn = date2num(dt)
    return dn


def snr2arcs(station,snrdata, azilims, elvlims, rhlims, precision, year,doy,signal='L1', normalize=False, 
        snrfigs=False, lspfigs=False, polydeg=2, gaptlim=5*60, pktnlim=4, 
        savefile=False, screenstats=False, l2c_only=False,satconsts=['G','R','E'], **kwargs):
    """
    reads an array of snr data (output from readklsnrtxt) and organises into:
    reflector height estimates, stats and detrended snr data for inverse analysis


    Parameters
    ----------
    station : str
        4 ch station name
    snrdata: numpy array 
        contents of SNR datafile

    azilims: list of floats
        azimuth angle limits (e.g., [90, 270])

    elvlims: list of floats
        elevation angle limits (e.g., [5, 30])

    rhlims: list of floats
        upper and lower reflector height limits (in metres) for quality control

    signal: str
        default 'L1' (C/A), can also use L2...if want to use L5 or whatever else you need to make some edits

    normalize: bool
        if you want to normalize the arcs so that they have the same amplitude

    snrfigs: bool
        if you want to produce some figures of SNR arcs

    lspfigs: bool
        if you want to produce some figures of Lomb-Scargle Periodograms

    polydeg: float
        degree of polynomial for DC 

    gaptlim: float
        if there is a gap in time bigger than [gaptlim] seconds in a particular arc then it will be ignored

    pktnlim: float
        peak to noise ratio qc condition = the peak of the LSP / mean of LSP within the range [rhlims]

    savefile: bool
        if you want to save the output to a pickle file then use this parameter as the name (string)

    kwargs: see below

    tempres: int
        if want to use different temporal resolution to input data (in seconds)

    satconsts: default use all given, otherwise specify from ['G', 'R', 'E'] (gps / glonass / galileo)

    Returns
    -------
    rh_arr: numpy array 
        reflector height estimates and stats

    snrdt_arr: numpy array 
        detrended SNR data for inverse analysis

    """
    print('screen stats', screenstats)
    # get a list for late ron
    if 'l2c_only' in kwargs:
        l2c_only = kwargs.get('l2c_only')
    # no need to print to screen
    #if l2c_only:
    #    print('For L2, only use L2c ')

    # get the list
    l2clist,l5list = l2c_l5_list(year,doy)
    # convert SNR to linear scale (from dB-Hz)
    snrdata[:, 4] = 10 ** (snrdata[:, 4] / 20)
    if lspfigs:
        print('Lomb Scargle Plots are being stored in the plots subdirectory')
        print('Creating these slows down the code and should never be done for routine processing')
    if snrfigs:
        print('SNR Plots are being stored in the plots subdirectory')
        print('Creating these slows down the code and should never be done for routine processing')

    # this is a decimator.  
    if 'tempres' in kwargs:
        tempres = kwargs.get('tempres')
        if (tempres != 1):
            print('Decimating the SNR file')
            tfilter = np.where(np.mod(snrdata[:, 3] - snrdata[0, 3], tempres) == 0)[0]
            snrdata = snrdata[tfilter]
            print('Num of obs after decimation', len(snrdata))

    #print('sat consts before if statement', satconsts)
    # remove satellite data from non-requested constellations
    if 'satconsts' in kwargs:
        satconsts = kwargs.get('satconsts')
        print('Now the sat consts', satconsts)
    if 'G' not in satconsts:
        nogps = snrdata[:, 0] > 100
        snrdata = snrdata[nogps]
        print('Num of obs after GPS removal', len(snrdata))

    if 'R' not in satconsts:
        noglo = np.logical_or(snrdata[:, 0] < 100, snrdata[:, 0] > 200)
        snrdata = snrdata[noglo]
        print('Num of obs after Glonass removal', len(snrdata))
    if 'E' not in satconsts:
        nogal = np.logical_or(snrdata[:, 0] < 200, snrdata[:, 0] > 300)
        snrdata = snrdata[nogal]
        print('Num of obs after Galileo removal', len(snrdata))
    if 'C' not in satconsts:
        nogal = np.logical_or(snrdata[:, 0] < 300, snrdata[:, 0] > 400)
        snrdata = snrdata[nogal]
        #print('Num of obs after Beidou removal', len(snrdata))

    # apply elevation and azimuth restrictions.   
    # KL: we usually do the DC first and then apply the restrictions DC  
    print('Apply elevation angle filter: ', elvlims[0], elvlims[1])
    tfilter = np.logical_and(snrdata[:, 1] > elvlims[0], snrdata[:, 1] < elvlims[1])
    snrdata = snrdata[tfilter]
    if azilims[0] < azilims[1]:
        tfilter = np.logical_and(snrdata[:, 2] > azilims[0], snrdata[:, 2] < azilims[1])
    else:
        tfilter = np.logical_or(snrdata[:, 2] > azilims[0], snrdata[:, 2] < azilims[1])
    snrdata = snrdata[tfilter]
    print('Apply azimuth angle filter: ', azilims[0], azilims[1])

    #print('Using Lomb Scargle precision of ', precision, ' m')
    # setting up arrays for the output of the LSP
    # changed this to 12 to accommodate frequency
    #rh_arr = np.empty((0, 11))
    rh_arr = np.empty((0, 12))
    # kl added a column for frequency
    #snrdt_arr = np.empty((0, 4), dtype=object)
    snrdt_arr = np.empty((0, 5), dtype=object)
    # get rid of satellite 51, 46 because not sure what it is
    # KL - not necessary - there are legal satellites > 32 in galileo
    #nobadsats = np.logical_or(snrdata[:, 0] < 33, snrdata[:, 0] > 99)
    #snrdata = snrdata[nobadsats, :]

    print('Using this signal: ', signal)
    # speed of light should be a declared constant
    # this should be cheked before you call the code
    #if signal not in ['L1','L2','L5','L1+L2','L1+L5','L1+L2+L5','L1+L6','L6','L1+L2+L6','L1+L2+L5+L6']:
    #    print('code only currently works for L1, L2, L5, L6, and combinations therein- exiting')
    #    sys.exit()
    #print('Satellites:')
    satellite_list = np.unique(snrdata[:,0]); 
    #print(satellite_list)
    signal_list = signal2list(signal)
    #print('Frequencies: ', signal_list)
    # make a dictionary to keep track of the constellation/frequencies that are being used
    alld = {}
    # initialize values?
    kristine_dictionary(alld,'','')

    xdir = os.environ['REFL_CODE'] + '/Files/' + station + '/'

    print('invsnr lsp results written to :', xdir + 'my_lsp.txt')
    fout = open(xdir + 'my_lsp.txt', 'w+')
    for sat in satellite_list:
        # get the data for that satellite
        tfilter = (snrdata[:, 0] == sat)
        first_tempd = snrdata[tfilter]

        # added this so i can loop through more than one frequency.  hopefully
        for xsignal in signal_list:
            isignal = int(xsignal[1:2]) # integer version of frequency
            if screenstats:
                print('Working on satellite/signal:', int(sat),xsignal,isignal)

            # frequency (integer) should be in column "5"
            sfilter = (first_tempd[:, 5] == isignal)
            tempd = first_tempd[sfilter]

            # this is still the same ....  
            # use the constants in gps.py
            # 22feb09 add beidou
            if np.logical_or(sat < 100, np.logical_and(sat > 200, sat < 300)):
                if xsignal == 'L1':
                    lcar = g.constants.wL1
                elif xsignal == 'L2':
                    lcar = g.constants.wL2
                elif xsignal == 'L5':
                    lcar = g.constants.wL5
                # these are for galileo, should not have this in the GPS loop - 
                # but will be lazy here because of how it was originally written
                elif xsignal == 'L6':
                    lcar = g.constants.wgL6
                elif xsignal == 'L7':
                    lcar = g.constants.wgL7
            elif np.logical_and(sat > 100, sat < 200):
                lcar = glonasswlen(int(sat), xsignal)
            elif np.logical_and(sat > 300, sat < 400):# beidou
                if xsignal == 'L2':
                    lcar = g.constants.wbL2
                if xsignal == 'L6':
                    lcar = g.constants.wbL6
                if xsignal == 'L7':
                    lcar = g.constants.wbL7

            # this restricts to L2C satellite but only if requested.
            # this does not mean the file has L2C data in it however.  unfortunately
            # particularly useful for trimble L2 data
            badone = False
            if ((sat < 100) & (xsignal == 'L2')) and l2c_only:
                if sat not in l2clist:
                    badone = True
                    #print('will not use this signal', xsignal, int(sat))

            # build up a dictionary that keeps track of which specific 
            # constellations/frequencies are observed
            alld = kristine_dictionary(alld,sat,xsignal)

            maxf = 2 * (rhlims[0] + rhlims[1]) / lcar
            precisionf = 2 * precision / lcar  # 1 mm was what he set - this is now a variable
            if (not np.isnan(maxf)) and (not badone):
                f = np.linspace(precisionf, maxf, int(maxf / precisionf))
                tempd = tempd[tempd[:, 3].argsort()]  # sort by time
                elv_tosort = np.array(tempd[:, 1], dtype=float)
                date_tosort = np.array(tempd[:, 3], dtype=float)
                ddate = np.ediff1d(date_tosort)
                delv = np.ediff1d(elv_tosort)
                bkpt = len(ddate)

                bkpt = np.append(bkpt, np.where(ddate > gaptlim)[0])  # gaps bigger than gaptlim
                if screenstats:
                    print('gaps',len(bkpt))
                bkpt = np.append(bkpt, np.where(np.diff(np.sign(delv)))[0])  # elevation rate changes direction
                if screenstats:
                    print('gaps from ostensible elevation rate changes',len(bkpt))
                bkpt = np.unique(bkpt)
                bkpt = np.sort(bkpt)
                if screenstats:
                    print('Number of arcs for this satellite?', len(bkpt))
                for ii in range(len(bkpt)):
                    if ii == 0:
                        sind = 0
                    else:
                        sind = bkpt[ii - 1] + 1
                    eind = bkpt[ii] + 1
                    if eind - sind < 20:
                        #print('arc not big enough')
                        # should this really be continue?
                        continue
                    elvt = np.array(tempd[sind:eind, 1], dtype=float)
                    if len(np.unique(elvt)) == 1:
                        print('unchanging elevation')
                        # should this really be continue?
                        continue
                    azit = np.array(tempd[sind:eind, 2], dtype=float) # azimuth array
                    sinelvt = np.sin(elvt / 180 * np.pi) # sine elevation angle array
                    datet = np.array(tempd[sind:eind, 3], dtype=float) # seconds in GPSish time?
                    snrt = np.array(tempd[sind:eind, 4], dtype=float) # snr data (dc removed) 
                # here dave is removing the direct signal
                    z = np.polyfit(sinelvt, snrt, polydeg)
                    p = np.poly1d(z)
                # estimating the periodogram used to compute the dominant reflector 
                # frequency, and thus RH
                    snrdt = snrt - p(sinelvt)
                    pgram = LombScargle(sinelvt, snrdt, normalization='psd').power(f)
                    pgram = 2 * np.sqrt(pgram/len(sinelvt))
                # converting it into the proper units of RH (meters)
                    reflh = 0.5 * f * lcar
                # this is a very simplistic outlier detector
                    tfilter = np.logical_and(reflh > rhlims[0], reflh < rhlims[1])
                    pgram_sub = pgram[tfilter]
                    reflh_sub = reflh[tfilter]
                    maxind = np.argmax(pgram_sub)
                # KL pktn = np.max(pgram_sub) / np.mean(pgram)
                    pktn = np.max(pgram_sub) / np.mean(pgram_sub)


                    maxF, maxA,peak2noise = simpleLSP(rhlims, lcar, precision,
                            elvt, sinelvt, snrdt,sat,xsignal,screenstats,fout,pktnlim)


                    #print('new', pktn, 'limit',pktnlim, 'old',np.max(pgram_sub) / np.mean(pgram))
                    if maxind != 0 and maxind != len(pgram_sub) - 1 and pktn > pktnlim:  # no peaks at either end of window
                        # KL moved this to a function/added a column to save frequency information
                        temp_arr = save_lsp_results(datet,maxind,reflh_sub,sat,elvt,azit,pgram_sub,snrdt,pktn,isignal)
                        rh_arr = np.vstack((rh_arr, temp_arr))
                        if normalize:
                            snrdt = snrdt * 100 / (np.max(np.abs(snrdt)))
                        satt = np.empty((len(datet)), dtype=int)
                        satt[:] = sat
                        nr = len(datet)
                        onecol = isignal*np.ones((nr,1))
                        # i added an integer column for frequencies
                        #temp_arr = np.column_stack([datet, satt, sinelvt, snrdt])
                        temp_arr2 = np.column_stack([datet, satt, sinelvt, snrdt, onecol])
                        # stack em
                        snrdt_arr = np.vstack((snrdt_arr, temp_arr2))
                        # and plots moved to function to clean this up
                        aa = str( int( np.mean(azit)) )
                        arc_plots(lspfigs, snrfigs, reflh,pgram,sat,datet,elvlims,elvt,snrdt,aa,xsignal)
    rh_arr = rh_arr[rh_arr[:, 0].argsort()]
    # make sure that arrays are sorted by time
    snrdt_arr = snrdt_arr[snrdt_arr[:, 0].argsort()]
    fout.close()
    if savefile:
        arcfilestr = 'arcsout.pkl'
        f = open(arcfilestr, 'wb')
        pickle.dump(rh_arr, f)
        pickle.dump(snrdt_arr, f)
        f.close()
    return rh_arr, snrdt_arr, alld


def residuals_cubspl_spectral(kval, knots, rh_arr):
    """
    function needed for inverse analysis

    Parameters
    ----------
    kval : ?

    knots : numpy array

    rh_arr : numpy array
        reflector heights in meters

    """
    tfilter = np.logical_and(rh_arr[:, 0] >= knots[0], rh_arr[:, 0] <= knots[-1])
    rh_arr = rh_arr[tfilter]
    dt_even = 60
    t_even = np.linspace(knots[0], knots[-1], int((knots[-1] - knots[0]) / dt_even) + 1)
    cubspl_f = interpolate.interp1d(knots, kval, kind='cubic')
    cubspl_even = cubspl_f(t_even)
    dhdt_even = np.gradient(cubspl_even, dt_even)
    f = interpolate.interp1d(t_even, dhdt_even)
    tgpst = np.array(rh_arr[:, 0], dtype=float)
    reflh = np.array(rh_arr[:, 1], dtype=float)
    tane_dedt = np.array(rh_arr[:, 3], dtype=float)
    dhdt = f(tgpst)
    cubspl_adj = cubspl_f(tgpst) + dhdt * tane_dedt
    residual_spectral = cubspl_adj - reflh
    return residual_spectral


def residuals_cubspl_js(inparam, knots, satconsts, signal, snrdt_arr,final_list,Nfreq):
    """
    function needed for snr-fitting inverse analysis
    js must stand for joakim strandberg ???

    this has to be modified for multi-frequency
    fspecdict and Nfreq
    22feb09 added beidou

    Parameters
    ----------
    inparam : 

    knots : 

    satconsts :

    signal :

    snrdt_arr :

    final_list :

    Nfreq :
    """
    if len(inparam) - Nfreq * 2 == len(knots):
        # then no roughness
        rh_kval = inparam[: - Nfreq * 2]
        satparams = inparam[- Nfreq * 2:]
        rough_in = False
    elif len(inparam) - Nfreq * 2 - 1 == len(knots):
        # roughness
        rh_kval = inparam[: - Nfreq * 2 - 1]
        satparams = inparam[- Nfreq * 2 - 1: -1]
        rough_in = inparam[-1]
    else:
        print('issue with length of input parameter array')
        return
    tmpc = 0
    res = np.empty(0)
    xsignal = signal
    signal_list = signal2list(xsignal)

    # G,R,E in first column and frequency in the second
    for satc in final_list:
        # this is where tmpc was originally
        tmpc = tmpc + 1
        c = satc[0:1] # constellation: E,G,R
        f = int(satc[1:2]) # frequency: 1,2,5,6,7
        
        # figuring out which satellites are which
        if c == 'G':
            tfilter = snrdt_arr[:, 1] < 100
        elif c == 'R':
            tfilter = np.logical_and(snrdt_arr[:, 1] > 100, snrdt_arr[:, 1] < 200)
        elif c == 'E':
            tfilter = np.logical_and(snrdt_arr[:, 1] > 200, snrdt_arr[:, 1] < 300)
        elif c == 'C':
            tfilter = np.logical_and(snrdt_arr[:, 1] > 300, snrdt_arr[:, 1] < 400)
        else:
            print('unknown satellite constellation')
        # take out the constellation
        constellation_tmp_arr = snrdt_arr[tfilter, :]
        # this is no longer a loop
        # for xsignal in signal_list:
        if True:
            #isignal = int(xsignal[1:2]) # integer f, i.e. 1,2,5
            isignal = f
            # now further restrict to a frequency for a given constellation
            ii = (constellation_tmp_arr[:,4] == isignal)
            tmp_arr = constellation_tmp_arr[ii,:]

            # this code expects L1 and not just a 1
            lcar = satfreq2waveL(c, 'L' + str(f), tmp_arr[:,1])
            #lcar = satfreq2waveL(satc, xsignal,tmp_arr[:,1])

            tgpst = np.array(tmp_arr[:, 0], dtype=float)
            sin_et = np.array(tmp_arr[:, 2], dtype=float)
            snr_dtt = np.array(tmp_arr[:, 3], dtype=float)
            cubspl_f = interpolate.interp1d(knots, rh_kval, kind='cubic')
            bf = np.logical_and(tgpst >= np.min(knots), tgpst <= np.max(knots))
            tgpst = tgpst[bf]
            sin_et = sin_et[bf]
            snr_dtt = snr_dtt[bf]

            h1 = cubspl_f(tgpst)
            modelsnr = satparams[int((tmpc - 1)*2)] * np.sin(4 * np.pi * h1 * sin_et / lcar) + satparams[int((tmpc - 1)*2 + 1)] * np.cos(4 * np.pi * h1 * sin_et / lcar)

            if rough_in:
                lk = 2 * np.pi / lcar
                modelsnr = modelsnr * np.exp(-4 * lk ** 2 * rough_in * sin_et ** 2)
            res = np.append(res, modelsnr - snr_dtt)
    return res



def snr2spline(station,year,doy, azilims, elvlims,rhlims, precision, kdt, snrfit=True, signal='L1', savefile=False, doplot=True, rough_in=0.1, **kwargs):
    """
    function analyzes a SNR file and outputs a fitted spline

    note that the file must be 24 hours long or it will not work

    Parameters
    ----------
    station : str
        4 ch station name
    year : int
        full year

    doy : int
        day of year

    azilims: list of floats
        azimuth angle limits (e.g., [90, 270])

    elvlims: list of floats
        elevation angle limits (e.g., [5, 30])

    rhlims: list of floats
        upper and lower reflector height limits (in metres) for quality 
        control e.g., [5, 10] is 5 and 10 m

    precision : float
        precision of the periodogram (m) 

    kdt: float
        spline knot spacing in seconds

    knots are spaced evenly except for at the start and end of the day
    the idea is that you could piece together the outpt from different days to have a continuous spline
    if kdt = 2 * 60 * 60 (2 hours), then knots at 0h, 1h, 3h,... 21h, 23h, 24h
    The idea is that you would ignore the first and last knots and then you could have a continuous spline
    with knots every 2 hours over multiple days
    snrfit: True or False if you want to do inverse modelling of the SNR data

    signal: 'L1', 'L2', currently under development

    savefile: set True if you want to save the output to a file

    doplot: set True if you want to produce a plot with the output from the analysis

    rough_in: 'roughness' parameter in the inverse modelling of SNR data (see Strandberg et al., 2016)

    kwargs: see below

    tempres: if want to use different temporal resolution to input data (in seconds)

    satconsts: default use all given, otherwise specify from ['G', 'R', 'E'] (gps / glonass / galileo)

    Returns
    -------
    invout: dictionary 
        outputs from inverse analysis

    This documentation was provided by the original author, David Purnell
    """

    #print(doplot, ' doplot setting')
    imodel = 0 # no refraction
    if 'lsp' in kwargs:
        lsp = kwargs.get('lsp')
        print('Thisi is a minimal refraction correction. Please submit a PR if you')
        print('would like to implement a better one.')
        if lsp['refraction']:
            imodel = 1 #
        # this should be the correct date if you want a better model 
        dmjd = 59580 # fake number
        p,T,irefr,humidity, Tm, lapse_rate = set_refraction_model(station, dmjd,lsp,imodel)
        #print('refraction parameters ', p,T,humidity,irefr)

    if 'rough_in' in kwargs:
        rough_in  = kwargs.get('rough_in')

    risky = False
    if 'risky' in kwargs:
        risky = kwargs.get('risky')

    lastday_seconds = 0
    if 'lastday_seconds' in kwargs:
        lastday_seconds = kwargs.get('lastday_seconds')

    if risky: 
        print('You are a risk taker.')
    else:
        print('You are not a risk taker.')

    no_dots = False
    if 'no_dots' in kwargs:
        no_dots = kwargs.get('no_dots')

    if 'screenstats' in kwargs:
        screenstats = kwargs.get('screenstats')

    if 'outlier_limit' in kwargs:
        outlier_limit = kwargs.get('outlier_limit')

    # default is 66 - originally I think david had it as a string?  
    snr_ending = 66
    if 'snr_ending' in kwargs:
        snr_ending = kwargs.get('snr_ending')
    
    if 'doy_end' in kwargs:
        doy_end = kwargs.get('doy_end')
    else:
        doy_end = doy


    if (doy == doy_end):
        print('Only doing one day')
        snrfile, snrdir, stryear, strdoy = define_inputfile(station,year,doy,snr_ending)
        snrdata = readklsnrtxt(snrfile, snrdir,signal)
        numdays = 1
    else:
        numdays = doy_end - doy + 1
        for d in range(doy, doy_end+1):
            snrfile, snrdir, stryear_temp, strdoy_temp = define_inputfile(station,year,d,snr_ending)
            snrtemp = readklsnrtxt(snrfile, snrdir,signal)
            if d == doy:
                snrdata = snrtemp
                # save these
                stryear = stryear_temp
                strdoy = strdoy_temp
            else:
                snrdata = np.vstack((snrdata, snrtemp))

    # set output directory here
    g.set_subdir(station)
    xdir = os.environ['REFL_CODE'] + '/Files/' + station + '/'


    nr,nc = snrdata.shape
    print('Number of obs', nr, ' number of days', numdays)
    if imodel == 1:
        #print('Modifying elevation angles for a simple refraction correction')
        #print('Elevation angle before ', snrdata[0,1])
        correctedE = refr.corr_el_angles(snrdata[:,1], p,T)
        snrdata[:,1] = correctedE
    #print('Elevation angle after', snrdata[0,1])

    invout = {}
    # starting point - which is taken from the first day 
    stryday = stryear + ':' + strdoy + ':00:00:00'

    print(stryday)
    tobj = Time(stryday, format='yday')
    gbase = tobj.gps
    #print('>>>>  gbase value', gbase)
    # Setting up knots ...
    # now deal with a final file that isn't 24 hours long ... 
    if (lastday_seconds == 0):
        total_seconds = numdays*86400
    else:
        total_seconds = (numdays-1)*86400 + lastday_seconds
    print('Total seconds ', total_seconds)
    knots = np.linspace(gbase + int(kdt/2), gbase + total_seconds - int(kdt/2), int(total_seconds/kdt))
    #print('Here be the knots:', knots)
    print('Knot spacing in seconds ',int(kdt/2))
    print('Number of knots',int(total_seconds/kdt))
    knots = np.append(gbase, knots)  # add start and end of day for more stable output but dont use these points
    knots = np.append(knots, gbase + total_seconds)


    #print('Sorting snr data into arcs')

    # arguments sent directly
    print('Begin Lomb Scargle analysis')
    s1=time.time()
    rh_arr, snrdt_arr, fspecdict= snr2arcs(station,snrdata, azilims, elvlims, rhlims, precision, year, doy, signal=signal,**kwargs)
    # should already exist
    #xdir = os.environ['REFL_CODE'] + '/Files/'
    #if not os.path.isdir(xdir):
    #    print('make output directory for file')
    #    subprocess.call(['mkdir',xdir])
    s2=time.time()
    print('Time spent: ',round(s2-s1,2), ' seconds')
    #print(fspecdict)
    print('Found ' + str(np.ma.size(rh_arr, axis=0)) + ' arcs')
    if screenstats:
        print(' RH(m)  Sat  Azim  Dt  Pk2noise Freq  Hours since epoch0')
        for i in range(0,len(rh_arr)):
            print(" %5.2f %3.0f %5.1f %5.1f %5.1f  %1.0f %10.3f " % ( 
                rh_arr[i,1], rh_arr[i,2], rh_arr[i,6], rh_arr[i,9]/60, rh_arr[i,10], rh_arr[i,11],(rh_arr[i,0]-gbase)/3600 ))
    if np.ma.size(rh_arr, axis=0) == 0:
        print('no reflector height data - exit')
        exit()

    if 'satconsts' in kwargs:
        satconsts = kwargs.get('satconsts')
    else:
        allsats = np.unique(rh_arr[:, 2])
        satconsts = []
        if len(np.where(np.logical_and(allsats > 0, allsats < 100))[0]) > 0:
            satconsts = np.append(satconsts, 'G')
        if len(np.where(np.logical_and(allsats > 100, allsats < 200))[0]) > 0:
            satconsts = np.append(satconsts, 'R')
        if len(np.where(np.logical_and(allsats > 200, allsats < 300))[0]) > 0:
            satconsts = np.append(satconsts, 'E')
        if len(np.where(np.logical_and(allsats > 300, allsats < 400))[0]) > 0:
            satconsts = np.append(satconsts, 'C')
    if np.ma.size(rh_arr, axis=0) < 2:
        print('not enough data - exit')
        exit()

    # sort the results by time??
    temp_dn = np.sort(rh_arr[:, 0])
    temp_dn = np.append(gbase, temp_dn)
    #temp_dn = np.append(temp_dn, gbase + numdays*86400)
    temp_dn = np.append(temp_dn, gbase + total_seconds)
    maxtgap = np.max(np.ediff1d(temp_dn))
    mintgap = np.min(np.ediff1d(temp_dn))
    if mintgap < 0:
        print('issue - values not in order')
        exit()
    print('max gap is ' + str(int(maxtgap / 60)) + ' minutes')

    if maxtgap > kdt * 1.05:  # giving 5% margin?
        print('Gap in data bigger than node spacing, which has risk of instabilities.')
        if (not risky):
            print('If you are a risk-taker, you need to set -risky True and rerun the code.')
            sys.exit()

    def residuals_spectral_ls(kval):
        residuals = residuals_cubspl_spectral(kval, knots, rh_arr)
        return residuals

    print('Now fitting a cubic spline to the arcs, but DP does not include rhdot')
    kval_0 = np.nanmean(rh_arr[:, 1]) * np.ones(len(knots))
    #print('Number of knots used in ', len(knots))
    s1=time.time()
    ls_spectral = least_squares(residuals_spectral_ls, kval_0, method='trf', bounds=rhlims)
    kval_spectral = ls_spectral.x
    print('Length of kval_spectral', len(kval_spectral))
    invout['knots'] = knots
    invout['kval_spectral'] = kval_spectral[1:-1]  # dont save first and last points
    s2=time.time()
    print('Fitting spline to LSP results took ', round(s2-s1,2), ' seconds')
    print('satellite constellations ', satconsts)
    if snrfit:
        s1=time.time()
        #print('Now doing Joakim Strandberg SNR fitting inversion')
        kval_0 = kval_spectral
        #print('kval_0', kval_0) dont need to print htis out
        final_list, Nfreq = smarterWay(fspecdict)
        print(final_list)
        #print('Number of constellation specific frequencies', Nfreq)
        # consts = len(satconsts)
        #kval_0 = np.append(kval_0, np.zeros(consts * 2))
        # this should be correct .... wont 
        kval_0 = np.append(kval_0, np.zeros(Nfreq* 2))
        #print('Roughness', rough_in)
        kval_0 = np.append(kval_0, rough_in)

        aa,bb = snrdt_arr.shape
        #print('Dimensions of the snrdt_arr variable', aa,bb)
        #print('Now sending it ', Nfreq, ' different constellation specific frequencies')
        def residuals_js_ls(inparam):
            residuals = residuals_cubspl_js(inparam, knots, satconsts, signal, snrdt_arr,final_list,Nfreq)
            return residuals
        #print('Calling the least squares code')
        ls_js = least_squares(residuals_js_ls, kval_0, method='lm')
        invout_js = ls_js.x
        kval_js = invout_js[:len(knots)]
        outparams_js = invout_js[len(knots):]
        invout['kval_js'] = kval_js[1:-1]  # dont save first and last points
        invout['outparams_js'] = outparams_js
        s2 = time.time()
        print('Time spent in snrfit', round(s2-s1,2), ' seconds')

    if True:
        delta_out = 300 # seconds
        if 'delta_out' in kwargs:
            delta_out = int(kwargs.get('delta_out'))
        outfile_type= 'txt'
        if 'outfile_type' in kwargs:
            outfile_type = kwargs.get('outfile_type')
        outfile_name= ''
        if 'outfile_name' in kwargs:
            outfile_name = kwargs.get('outfile_name')
        #plotdt = 5 * 60
        plotdt = delta_out # make it so people can change it
        #tplot = np.linspace(gbase, gbase + numdays*86400, int(86400/plotdt + 1))

        
        #numvals = 1 + int(numdays*86400/delta_out)
        numvals = 1 + int(total_seconds/delta_out)

        #nv2 = 1 + int(numdays*86400/3600)
        nv2 = 1 + int(total_seconds/3600)
        tplot = np.linspace(gbase, gbase + total_seconds, numvals)

        tplot_hourly = np.linspace(gbase, gbase + total_seconds, nv2 )
        lsp_per_hour = []
        for ijk in range(0,len(tplot_hourly)): 
            H1 = float((tplot_hourly[ijk]-gbase)/3600) ; H2 = H1 + 1
            dt = (rh_arr[:,0] - gbase)/3600
            ii = np.logical_and(dt>= H1, dt< H2)
            lsp_per_hour.append(len(rh_arr[ii,1]))
            #print(H1, len(rh_arr[ii,1]))
        tplot_dn = gps2datenum(tplot)
        cubspl_f = interpolate.interp1d(knots, kval_spectral, kind='cubic')
        rh_spectral_plot = cubspl_f(tplot)
        fig, ax = plt.subplots(figsize=(8, 4))
        rh_dn = gps2datenum(np.array(rh_arr[:, 0], dtype=float))
        if not (no_dots):
            plot_tracks(rh_arr, rh_dn)
            pspec, = plt.plot_date(tplot_dn, rh_spectral_plot, '-.',color='gray')
            pspec.set_label('cubspl')

        if snrfit:
            cubspl_f = interpolate.interp1d(knots, kval_js, kind='cubic')
            rh_js_plot = cubspl_f(tplot)
            #spline_at_lsp = cubspl_f(rh_arr[:,0]) does not work - doesn't like objects
            # maybe this will work ... 
            Xres = np.empty(shape=[0,1])
            # no doubt there is an easier way ...
            for ijk in range(0,len(rh_arr)):
                tii = float(rh_arr[ijk,0]) # change from object?
                Xres = np.append(Xres,  tii)

            spl_at_lsp = cubspl_f(Xres)
            ab = rh_arr

            print('invsnr lsp outliers written to : ', xdir + 'outliers_lsp.txt')
            iiout = open(xdir + 'outliers_lsp.txt', 'w+')
            for ijk in range(0,len(rh_arr)):
                dd =  rh_arr[ijk,1]-spl_at_lsp[ijk]
                if (abs(dd) > outlier_limit):
                    iiout.write("{0:5.2f} {1:3.0f} {2:4.1f} {3:5.2f} \n".format(dd, ab[ijk,2], ab[ijk,6], ab[ijk,10]))
            iiout.close()

            # opens the file, writes a header
            iout, usetxt = invsnr_header(xdir, outfile_type, station, outfile_name) 
            
            # remove last point ...
            for ijk in range(0,len(rh_js_plot)-1):
                # undo dave's time units (rel gps) into a datetime object
                # which hour from start time does this belong to?
                whichhour = math.floor( (tplot[ijk] - gbase)/3600)
                #print(whichhour,lsp_per_hour[whichhour])
                dt = gps2datetime(tplot[ijk])
                y= dt.year; m=dt.month; d=dt.day; h=dt.hour; mi = dt.minute ; s=int(dt.second)
                # get doy = i am sure this can be done better but i am not fluent in datetime
                doy, cdoy, cyyyy, cyy = g.ymd2doy(y,m,d)
                # get the MJD value
                MJD, fracS = g.mjd(y,m,d,h,mi,s)
                #print(y,m,d,h,mi,s,doy,MJD+fracS, rh_js_plot[ijk])
                if usetxt:
                    iout.write(" {0:4.0f} {1:2.0f} {2:2.0f} {3:2.0f} {4:2.0f} {5:2.0f} {6:8.3f} {7:3.0f} {8:13.6f} {9:3.0f}\n".format(y, m, 
                        d,h,mi,s,rh_js_plot[ijk],doy,MJD+fracS, lsp_per_hour[whichhour]))
                else:
                    iout.write(" {0:4.0f},{1:2.0f},{2:2.0f},{3:2.0f},{4:2.0f},{5:2.0f},{6:8.3f},{7:3.0f},{8:13.6f},{9:3.0f}\n".format(y, m, 
                        d,h,mi,s,rh_js_plot[ijk],doy,MJD+fracS),lsp_per_hour[whichhour])
            pjs, = plt.plot_date(tplot_dn, rh_js_plot, '-',color='black')
            pjs.set_label('invmod')
            iout.close()

        dformat = DateFormatter('%Y-%m-%d')
        ax.xaxis.set_major_formatter(dformat)
        ax.set_title('Station ' + station.upper() + ' ' + signal)
        ax.set_xlim(gps2datenum(gbase), gps2datenum(gbase + total_seconds))
        #ax.set_xlim(gps2datenum(gbase), gps2datenum(gbase + numdays*86400))
        ax.set_xticks(np.linspace(gps2datenum(gbase), gps2datenum(gbase + total_seconds), int(86400 / (60 * 60 * 6) + 1)))
        #ax.set_xticks(np.linspace(gps2datenum(gbase), gps2datenum(gbase + numdays*86400), int(86400 / (60 * 60 * 6) + 1)))
        ax.xaxis.set_major_formatter(DateFormatter('%m-%d %H:%M'))
        #ax.xaxis.set_major_formatter(DateFormatter('%Y-%m-%d %H:%M'))
        ax.legend(loc="upper right", prop={"size":8})
        #ax.legend(loc="upper right",bbox_to_anchor=(1. , 0.7),prop={"size":8})
        #plt.xticks(rotation =45); 
        ax.grid(True)
        ax.set_ylabel('RH meters')
        ax.invert_yaxis()
        spng = str(os.environ['REFL_CODE']) + '/Files/' + station + '/' + 'spline_out.png'
        print('Plot with results has been written to ', spng)
        plt.savefig(spng)
        if doplot:
            plt.show()
            plt.close()

    if savefile:
        invfilestr = snrfile + '.inv'
        f = open(invfilestr, 'wb')
        pickle.dump(invout, f)
        pickle.dump(rh_arr, f)
        f.close()
        print('dumped a pickle')
    return invout

def invsnr_header(xdir, outfile_type,station,outfile_name):
    """
    Makes header for output of invsnr analysis

    Parameters
    ----------
    xdir : str
        directory for the output file
    outfile_type : str
        csv or txt
    station : str
        4 character name
    outfile_name : str
        name of output - if empty string, it uses default

    Returns 
    -------
    fileID : file
        used for writing to file
    usetxt : bool
        boolean for the code calling this function to use
        if you write out special files, they go in the working directory

    """
    if outfile_type == 'txt':
        usetxt = True; 
        if (len(outfile_name) == 0):
            ioutputfile= xdir + station + '_invsnr.txt'
        else:
            ioutputfile= outfile_name + '.txt'

        iout = open(ioutputfile, 'w+')
        commentl = '%' # for now ... 
    else:
        usetxt = False; 
        if (len(outfile_name) == 0):
            ioutputfile= xdir + station + '_invsnr.csv'
        else:
            ioutputfile= outfile_name + '.csv'

        iout = open(ioutputfile, 'w+')
        commentl = '%'
    print('invsnr output will be written to: ', ioutputfile)
    xxx = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
    iout.write('{0:s} Results for {1:4s} calculated on {2:20s} \n'.format(commentl,  station, xxx ))
    iout.write('{0:s} gnssrefl, https://github.com/kristinemlarson \n'.format(commentl))
    iout.write('{0:s} YYYY MM DD HH MM SS   RH(m) doy  MJD  Nretrievals\n'.format(commentl))

    return iout, usetxt

def define_inputfile(station,year,doy,snr_ending):
    """

    Parameters
    ----------
    station : str
        4 ch name of station 
    year : integer

    doy : int
        day of year 

    snr_ending : int
        file ending, e.g. 66, 99

    Returns
    -------
    snrfile: str
        name of snrfile 

    snrdir : str
        name of output directory

    cyyyy : str
        four character year

    cdoy : str
        three character day of year
    """
    cdoy = '{:03d}'.format(doy) ;
    cyyyy = str(year)
    cyy = cyyyy[2:4]

    xdir = os.environ['REFL_CODE'] + '/' + cyyyy + '/snr/' + station + '/'
    snrfile = station + cdoy + '0.' + cyy + '.snr' + str(snr_ending)
    gzfile = xdir + snrfile + '.gz'
    xzfile = xdir + snrfile + '.xz'
    snrdir = ''
    if not os.path.isfile(snrfile):
        #print('SNR file does not exist in the local directory')
        if os.path.isfile(xdir + snrfile):
            #print('SNR file does exist in the REFL_CODE directory')
            snrdir = xdir
        # look for gzipped file
        elif os.path.isfile(gzfile):
            # unzip it, declare success
            #print('Found gzipped version', snrfile)
            subprocess.call(['gunzip',gzfile])
            snrdir = xdir
        elif os.path.isfile(xzfile):
            # unxz it, declare success
            #print('Found xz version', snrfile)
            subprocess.call(['unxz',xzfile])
            snrdir = xdir
        else:
            print(snrfile, ' not found. Exiting')
            sys.exit()

    return snrfile,snrdir, cyyyy, cdoy

def arc_plots(lspfigs, snrfigs, reflh,pgram,sat,datet,elvlims,elvt,snrdt,azdesc,xsignal):
    """
    moved these individual plots out of the way

    lspfigs : bool

    snrfigs : bool

    reflh : numpy array
        reflector heights (m)

    pgram : numpy array
        periodogram ? 

    sat : numpy array 

    datet : datetime

    elvlims : list of floats
        min and max elev angle (deg)

    xsignal : str
        i think this is L1, L2 etc

    """
    if not os.path.isdir('plots'):
        print('make output directory for plots')
        subprocess.call(['mkdir','plots'])

    resol = 150
    nm = 'sat' + str(int(sat)) + '_' + xsignal + '_' + gps2datetime(np.mean(datet)).strftime('%H-%M')  
    if lspfigs:
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.plot(reflh, pgram)
        ax.grid(True)
        ax.set_xlabel('RH (m)')
        ax.set_title('Satellite ' + str(int(sat)) + ' Azim:' + azdesc)
        ax.set_ylabel('unknown power units')
        plt.savefig('plots/' + nm + '_LSP.png',dpi=resol)
        plt.close(fig)
    if snrfigs:
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.plot(elvt, snrdt)
        plt.xlim(elvlims[0], elvlims[1])
        ax.set_title('Satellite ' + str(int(sat)) + ' Azim:' + azdesc)
        #plt.ylim(-100, 100)
        ax.grid(True)
        ax.set_xlabel('elevation angle (deg)')
        ax.set_ylabel('volts/volts?')
        plt.savefig('plots/' + nm + '_SNR.png',dpi=resol)
        plt.close(fig)

def signal2list(signal):
    """
    turns signal input (e.g. L1+L2) to a list
    22feb09 tried to add more frequencies ...

    Returns
    -------
    signal_list : str

    """
    if (len(signal) == 2):
        signal_list = [signal]
    elif (len(signal) == 5):
        signal_list = [signal[0:2],signal[3:5]]
    elif (len(signal) == 8):
        signal_list = [signal[0:2],signal[3:5], signal[6:8]] 
    elif (len(signal) == 11):
        signal_list = [signal[0:2],signal[3:5], signal[6:8],signal[9:11] ]
    elif (len(signal) == 14):
        signal_list = [signal[0:2],signal[3:5], signal[6:8],signal[9:11], signal[12:14] ]

    return signal_list


def satfreq2waveL(satc, xsignal,fsatnos):
    """
    given satellite constellation ('G', 'E' ...)
    xsignal ('L1','L2' ...)
    satnos (satellite numbers)
    2022feb09 added Beidou.
    """
    if (satc == 'G'):
        if (xsignal == 'L1'):
            lcar = g.constants.wL1
        elif (xsignal == 'L2'):
            lcar = g.constants.wL2
        elif (xsignal == 'L5'):
            lcar = g.constants.wL5
    elif satc == 'R':
        if xsignal == 'L5':
            lcar = np.nan
        else:
            satnos = np.array(fsatnos, dtype=int)
            lcar = glonasswlen(satnos, xsignal)
    elif (satc == 'E'):
        if xsignal == 'L1':
            lcar = g.constants.wL1
        elif xsignal == 'L2':
            lcar = np.nan
        elif xsignal == 'L5':
            lcar = g.constants.wL5
        # added galileo l6,l7
        elif xsignal == 'L6':
            lcar = g.constants.wgL6
        elif xsignal == 'L7':
            lcar = g.constants.wgL7
    elif (satc == 'C'):
        if xsignal == 'L2':
            lcar = g.constants.wbL2
        if xsignal == 'L6':
            lcar = g.constants.wbL6
        if xsignal == 'L7':
            lcar = g.constants.wbL7

    return lcar

def loadsnrfile(snrfile, thedir):
    """
    loads the snr file , but does not pick out the signal.
    using two functions will make it easier to use more than one frequency


    do time modification here now. column 4 is time since GPS began, in seconds

    Parameters
    ----------
    snrfile : str
        name of the SNR file
    thedir : str
        location of the SNR file

    Returns
    -------
    snrdata : numpy array
        floats. Time (python col 3) is converted to fake gps time

    """
    snrin = thedir + snrfile

    print('Reading file:', snrin)
    # this assumes someone has checked existence first
    snrdata = np.loadtxt(snrin)

    stryear = str(int(snrfile[9:11]) + 2000)
    strdoy = snrfile[4:7]
    stryday = stryear + ':' + strdoy + ':00:00:00'
    tobj = Time(stryday, format='yday')
    gbase = tobj.gps
    # put the time tags into fake GPS time (seconds since GPS began?)
    snrdata[:, 3] = snrdata[:, 3] + gbase

    return snrdata

def plot_tracks(rh_arr, rh_dn):
    """
    send the array of LSP results (rh_arr) with time variable for 
    plotting (rh_dn)
    kl feb09 adding beidou

    Parameters
    ----------
    rh_arr : numpy array
        data used by inverse code. Need to add desc
    rh_dn : numpy array
        data used by inverse code. Need to add desc

    """
    ms=4
    ii = (rh_arr[:,2] < 100) & (rh_arr[:,11] == 1)# GPS
    ii2 = (rh_arr[:,2] < 100) & (rh_arr[:,11] == 2)# GPS
    ii5 = (rh_arr[:,2] < 100) & (rh_arr[:,11] == 5)# GPS

    jj = (rh_arr[:,2] > 100) & (rh_arr[:,2] < 200)  & (rh_arr[:,11] == 1) # Glonass
    jj2 = (rh_arr[:,2] > 100) & (rh_arr[:,2] < 200)  & (rh_arr[:,11] == 2) # Glonass

    kk = (rh_arr[:,2] > 200) & (rh_arr[:,2] < 300) & (rh_arr[:,11] == 1)# galileo
    kk5 =(rh_arr[:,2] > 200) & (rh_arr[:,2] < 300) & (rh_arr[:,11] == 5) # galileo
    kk6 =(rh_arr[:,2] > 200) & (rh_arr[:,2] < 300) & (rh_arr[:,11] == 6) # galileo
    kk7 =(rh_arr[:,2] > 200) & (rh_arr[:,2] < 300) & (rh_arr[:,11] == 7) # galileo

    mm = (rh_arr[:,2] > 300) &  (rh_arr[:,11] == 2)# beidou 
    mm6 = (rh_arr[:,2] > 300) & (rh_arr[:,11] == 6) # beidou
    mm7 = (rh_arr[:,2] > 300) & (rh_arr[:,11] == 7) # beidou

    if len(rh_dn[mm]) > 0:
        psec, = plt.plot_date(rh_dn[mm], rh_arr[mm, 1], '<',color='magenta',markersize=ms)
        psec.set_label('BDS L2')
    if len(rh_dn[mm6]) > 0:
        psec, = plt.plot_date(rh_dn[mm6], rh_arr[mm6, 1], '+',color='magenta',markersize=ms)
        psec.set_label('BDS L6')
    if len(rh_dn[mm7]) > 0:
        psec, = plt.plot_date(rh_dn[mm7], rh_arr[mm7, 1], 'v',color='magenta',markersize=ms)
        psec.set_label('BDS L7')

    if len(rh_dn[ii]) > 0:
        psec, = plt.plot_date(rh_dn[ii], rh_arr[ii, 1], 'o',color='blue',markersize=ms)
        psec.set_label('GPS L1')

    if len(rh_dn[ii2]) > 0:
        psec, = plt.plot_date(rh_dn[ii2], rh_arr[ii2, 1], '<',color='blue',markersize=ms)
        psec.set_label('GPS L2')

    if len(rh_dn[ii5]) > 0:
        psec, = plt.plot_date(rh_dn[ii5], rh_arr[ii5, 1], 's',color='blue',markersize=ms)
        psec.set_label('GPS L5')

    if len(rh_dn[kk]) > 0:
        psec, = plt.plot_date(rh_dn[kk], rh_arr[kk, 1], 'o',color='orange',markersize=ms)
        psec.set_label('GAL L1')

    if len(rh_dn[kk5]) > 0:
        psec, = plt.plot_date(rh_dn[kk5], rh_arr[kk5, 1], 's',color='orange',markersize=ms)
        psec.set_label('GAL L5')

    if len(rh_dn[kk6]) > 0:
        psec, = plt.plot_date(rh_dn[kk6], rh_arr[kk6, 1], '+',color='orange',markersize=ms)
        psec.set_label('GAL L6')

    if len(rh_dn[kk7]) > 0:
        psec, = plt.plot_date(rh_dn[kk7], rh_arr[kk7, 1], 'v',color='orange',markersize=ms)
        psec.set_label('GAL L7')

    if len(rh_dn[jj]) > 0:
        psec, = plt.plot_date(rh_dn[jj], rh_arr[jj, 1], 'ro',markersize=ms)
        psec.set_label('GLO L1')

    if len(rh_dn[jj2]) > 0:
        psec, = plt.plot_date(rh_dn[jj2], rh_arr[jj2, 1], '<', color='red',markersize=ms)
        psec.set_label('GLO L2')

def kristine_dictionary(alld,sat,xsignal):
    """
    22feb09 added beidou
    """
    if len(alld) == 0:
        alld['G1'] = False
        alld['G2'] = False 
        alld['G5'] = False
        alld['E1'] = False
        alld['E2'] = False
        alld['E5'] = False
        alld['E6'] = False
        alld['E7'] = False
        alld['R1'] = False
        alld['R2'] = False
        alld['R5'] = False
        alld['C2'] = False
        alld['C6'] = False
        alld['C7'] = False
    else:
        if (sat < 100) and (xsignal == 'L1'):
            alld['G1'] = True
        if (sat < 100) and (xsignal == 'L2'):
            alld['G2'] = True
        if (sat < 100) and (xsignal == 'L5'):
            alld['G5'] = True
        if ((sat > 100) and (sat < 200)) and (xsignal == 'L1'):
            alld['R1'] = True
        if ((sat > 100) and (sat < 200)) and (xsignal == 'L2'):
            alld['R2'] = True

        if ((sat > 200) and (sat < 300))  and (xsignal == 'L1'):
            alld['E1'] = True
        if ((sat > 200) and (sat < 300))  and (xsignal == 'L5'):
            alld['E5'] = True
        # added more galileo
        if ((sat > 200) and (sat < 300))  and (xsignal == 'L6'):
            alld['E6'] = True
        if ((sat > 200) and (sat < 300))  and (xsignal == 'L7'):
            alld['E7'] = True

        if ((sat > 300) and (sat < 400))  and (xsignal == 'L2'):
            alld['C2'] = True
        if ((sat > 300) and (sat < 400))  and (xsignal == 'L6'):
            alld['C6'] = True
        if ((sat > 300) and (sat < 400))  and (xsignal == 'L7'):
            alld['C7'] = True

    return alld

def smarterWay(a):
    """
    just want to know how many true values there are in the a dictionary 
    and then write them to a list, as in ['G1','G2']
    sure to be a better way - but this works for now

    """
    i=0
    final_list = []
    for val in a:
        if a[val]:
            i=i+1
            final_list.append(val)
    return final_list, i

def freq_out(x,ofac,hifac):
    """
    inputs: x
    ofac: oversamping factor
    hifac
    outputs: two sets of frequencies arrays
    """
#
# number of points in input array
    n=len(x)
#
# number of frequencies that will be used
    nout=int(0.5*ofac*hifac*n)

    xmax = np.max(x)
    xmin = np.min(x)
    xdif=xmax-xmin
# starting frequency
    pnow=1.0/(xdif*ofac)
    pstart = pnow
    pstop = hifac*n/(2*xdif)
#
# simpler way
    pd = np.linspace(pstart, pstop, nout)
    return pd

def get_ofac_hifac(elevAngles, cf, maxH, desiredPrec):
    """
    computes two factors - ofac and hifac - that are inputs to the
    Lomb-Scargle Periodogram code.
    We follow the terminology and discussion from Press et al. (1992)
    in their LSP algorithm description.

    Parameters
    ----------
    elevAngles:  numpy array
        satellite elevation angles in degrees

    cf: float
        L-band wavelength/2  in meters

    maxH: float
        maximum LSP grid frequency in meters

    desiredPrec: float
        the LSP frequency grid spacing in meters 

    Returns
    -------
    ofac: float
        oversampling factor

    hifac: float
        high-frequency factor
    """
# in units of inverse meters
    X= np.sin(elevAngles*np.pi/180)/cf

# number of observations
    N = len(X)
# observing Window length (or span)
# units of inverse meters
    W = np.max(X) - np.min(X)

# characteristic peak width, meters
    cpw= 1/W

# oversampling factor
    ofac = cpw/desiredPrec

# Nyquist frequency if the N observed data samples were evenly spaced
# over the observing window span W, in meters
    fc = N/(2*W)

# Finally, the high-frequency factor is defined relative to fc
    hifac = maxH/fc

    return ofac, hifac

def simpleLSP(rhlims, lcar, precision,elvt, sinelvt, snrdt,sat,xsignal,screenstats,fout,pktnlim):
    """
    Parameters
    ----------
    input 

    rhlims from dave's code (rhmin and rhmax)
    lcar :  is gnss wavelength in m
    precision of the periodogram, in meters
    elvt - elevation angles in degrees
    sinelvt, sine elevation angle
    snrdt - detrended snr data
    """
    maxH = rhlims[-1] # meters
    cf = lcar/2 # wavelength/2

    # get oversampling factor and hifac (how far in RH space)
    ofac, hifac = get_ofac_hifac(elvt, cf, maxH, precision)
    # scaled elevation angle
    scaledE = sinelvt /cf
    px = freq_out(scaledE,ofac,hifac)
    scipy_LSP = spectral.lombscargle(scaledE, snrdt, 2*np.pi*px)
    pz = 2*np.sqrt(scipy_LSP/len(sinelvt))
    # impose rh limits
    ii = (px >= rhlims[0]) & (px <= rhlims[1])
    px = px[ii]
    pz = pz[ii]
    noise = np.mean(pz)
    # don't allow max to be at hte beginning of the end

    ij = np.argmax(pz)
    maxF = px[ij]
    maxAmp = np.max(pz)
    #print(ij, maxF,len(px))
    bad = False
    if int(ij) == (len(px)-1): # end
        bad = True
        maxAmp = np.nan; maxF = np.nan ; peak2noise = np.nan
    if int(ij) == 0: # beginning
        bad = True
        maxAmp = np.nan; maxF = np.nan ; peak2noise = np.nan

    peak2noise = maxAmp/noise
    if (peak2noise < pktnlim): 
        bad = True
        #print('fail pk2noise',peak2noise, pktnlim)
    # won't print for now.  too messy to have two sets
    # "{0:4.0f} {1:3.0f} {2:s} {3:s}  {4:s} {5:2.0f} {6:2.0f}
    if screenstats and (not bad):
        fout.write("{0:5.2f} {1:5.2f} {2:5.1f} {3:3.0f} {4:s} \n".format(maxF, maxAmp,peak2noise, sat, xsignal))
    return maxF, maxAmp, peak2noise

def l2c_l5_list(year,doy):
    """
    for given year and day of year, returns a satellite list of
    L2C and L5 transmitting satellites

    to update this numpy array, the data are stored in a simple triple of PRN number, launch year,
    and launch date.
    author: kristine larson
    date: march 27, 2021
    june 24, 2021: updated for SVN78

    this should point to gps.py

    """

    # this numpy array
    l2c=np.array([[1 ,2011 ,290], [3 ,2014 ,347], [4 ,2018 ,357], [5 ,2008 ,240],
        [6 ,2014 ,163], [7 ,2008 ,85], [8 ,2015 ,224], [9 ,2014 ,258], [10 ,2015 ,343],
        [11, 2021, 168],
        [12 ,2006 ,300], [14 ,2020 ,310], [15 ,2007 ,285], [17 ,2005 ,270],
        [18 ,2019 ,234], [23 ,2020 ,182], [24 ,2012 ,319], [25 ,2010 ,240],
        [26 ,2015 ,111], [27 ,2013 ,173], [29 ,2007 ,355], [30 ,2014 ,151], [31 ,2006 ,270], [32 ,2016 ,36]])
    # indices that meet your criteria
    ij=(l2c[:,1] + l2c[:,2]/365.25) < (year + doy/365.25)
    l2csatlist = l2c[ij,0]
    firstL5 = 2010 + 148/365.25 # launch may 28, 2010  - some delay before becoming healthy

    newlist = l2c[ij,:]
    ik= (newlist[:,1] + newlist[:,2]/365.25) > firstL5
    l5satlist = newlist[ik,0]

    return l2csatlist, l5satlist

def save_lsp_results(datet,maxind,reflh_sub,sat,elvt,azit,pgram_sub,snrdt,pktn,isignal):
    """
    just cleaning up - move the temp_arr definition to a function
    each column is defined below.

    Parameters
    ----------
    datet : float
        seconds in GPSish time 
    reflh_sub : numpy of floats?
        windowed rh estimates
    sat : int
        satellite number
    elvt : numpy array of floats 
        elevation angles(deg)
    azit : numpy array of floats
        azimuth angles (deg)
    snrdt : numpy array of floats
        detrended SNR data (DC component removed)
    pktn : float
        peak 2 noise via Dave Purnell's definition
    isignal : int
        frequency, 1,2, or 5

    Returns
    -------
    tmp_arr : numpy array (12 columns)

    """

    temp_arr = np.empty((1, 12), dtype=object)
    # saving some stats for each satellite arc
    temp_arr[0, 0] = int(np.round(np.mean(datet)))                          # time of arc in seconds?
    temp_arr[0, 1] = reflh_sub[maxind]                                      # reflector height (meters)
    temp_arr[0, 2] = sat                                                    # sat prn
    dthdt = ((elvt[-1] - elvt[0]) / 180 * np.pi) / (datet[-1] - datet[0])
    temp_arr[0, 3] = np.tan(np.mean(elvt) / 180 * np.pi) / dthdt            # tane / (de/dt)
    temp_arr[0, 4] = np.min(elvt)                                           # min elv (deg)
    temp_arr[0, 5] = np.max(elvt)                                           # max elv (deg)
    temp_arr[0, 6] = np.mean(azit)                                          # mean azimuth (deg)
    temp_arr[0, 7] = np.max(pgram_sub)                                      # peak of lsp
    temp_arr[0, 8] = np.var(snrdt)                                          # variance of lsp
    temp_arr[0, 9] = datet[-1] - datet[0]                                   # length (in time) of arc (seconds?)
    temp_arr[0, 10] = pktn                                                  # peak-to-noise ratio
    temp_arr[0, 11] = isignal                                               # frequency, integer

    return temp_arr


def set_refraction_model(station, dmjd,lsp,imodel):
    """
    imodel is 1 for simple refraction model
    eventually will add other refraction models

    Looks like this was copied from other code and should be
    consolidated ...

    Parameters
    ----------
    station : str
        4 ch station name
    dmjd : float
        modified julian date
    lsp : dictionary
        station information including latitude and longitude
    imodel : integer
        set to 1 (time varying off) or 0 (time varying on)

    Returns
    -------
    p : float
        pressure (units?)
    T : float
        temperature in deg C
    irefr : int
        number value written to output files to keep track of refraction model
    e : float
        water vapor pressure, hPa
    Tm : float
        temperature in kelvin
    lapse_rate : float
        see source code for details

    """
    xdir = os.environ['REFL_CODE']
    p = 0; T = 0; irefr = 0
    #print(lsp['lat'], lsp['lon'])
    if (imodel == 1):
        irefr = 1
        refr.readWrite_gpt2_1w(xdir, station, lsp['lat'], lsp['lon'])
# time varying is set to no for now (it = 1)
        it = 1
        # should use sea level, and this is ellipsoidal height.  so it is wrong
        dlat = lsp['lat']*np.pi/180; dlong = lsp['lon']*np.pi/180; ht = lsp['ht']
        p,T,dT,Tm,e,ah,aw,la,undu = refr.gpt2_1w(station, dmjd,dlat,dlong,ht,it)
        #print("Pressure {0:8.2f} Temperature {1:6.1f} \n".format(p,T))

    return p,T,irefr, e, Tm, la




# -*- coding: utf-8 -*-
from __future__ import division
import json
import numpy as np 
import os, datetime, traceback, gzip
import subprocess
import sys
from scipy.interpolate import interp1d

import gnssrefl.gps as g
import gnssrefl.decipher_argt as gt

def nmea_apriori_coords(station,llh,sp3):
    """
    """
    foundcoords = False
    recv = [0,0,0]
    # you input the coordinates on the command line
    if (llh[0] is not None) & (llh[1] is not None) & (llh[2] is not None):
        # compute Cartesian receiver coordinates in meters
        x,y,z = g.llh2xyz(llh[0],llh[1],llh[2])
        recv = [x,y,z]
        foundcoords = True

    if sp3 and (not foundcoords):
        # try to get the LLH from json file
        jfile1  = os.environ['REFL_CODE'] + '/input/' + station.upper() + '.json'
        jfile2  = os.environ['REFL_CODE'] + '/input/' + station.lower() + '.json'
        fexists = False
        if os.path.isfile(jfile1):
            jfile = jfile1 ; fexists = True
            print('json file found (using for station coordinates)',jfile1)
        elif os.path.isfile(jfile2):
            jfile = jfile2 ; fexists = True
            print('json file found (using for station coordinates)',jfile2)

        if fexists:
            with open(jfile, 'r') as my_json:
                fcont = json.load(my_json)

            llh[0] = fcont['lat']; llh[1] = fcont['lon'] ; llh[2] = fcont['ht']
            x,y,z = g.llh2xyz(llh[0],llh[1],llh[2])
            recv = [x,y,z]
            foundcoords = True
        else:
            print('This code requires lat/lon/ht inputs. Provide on the command line or')
            print('make a json file using gnssir_input.  Exiting')
            sys.exit()

    # if coordinates not found you are using low quality NMEA az el and thus do not need
    # a priori coordinates
    if foundcoords:
        print('Cartesian coordinates ', recv)

    return recv, foundcoords

#Last modified Feb 22, 2023 by Taylor Smith (git: tasmi) for additional constellation support

def nmea_translate(locdir, fname, snrfile, csnr, dec, year, doy, recv, sp3, gzip,orb,hour):
    """
    Reads and translates a NMEA file stored in locdir + fname.
    The naming convention assumed for the NMEA file is  SSSS1520.23.A
    where SSSS is station name, day of year is 152 and year is 2023
    locdir is generally $REFL_CODE/nmea/SSSS/yyyy where yyyy is the year number and SSSS is the station name

    Note from KL: I believe lowercase is also allowed (and preferred), but the A at the end is still set to be upper case
    (I believe) The SNR files are stored with upper case if given upper case, lower case if given lower case.

    At request of Felipe Nievinski, the ultra option (which points to GFZ ultrarapid orbits) will
    first look for the orbit on (requested doy, hour 0), then (doy-1, hour 0), and then (doy-1, hour 12).
    I do not think this first search is correct - but I don't think it makes the code crash so I will
    keep it there for now.

    Parameters
    -----------
    locdir : str
        directory where your NMEA files are kept
    fname : str
        NMEA filename 
    snrfile : str
        name of output file for SNR data
    csnr : str
        snr option, i.e. '66' or '99'
    dec : int
        decimation value in seconds
    year : int
        full year
    doy : int
        day of year
    recv : list of floats
        a priori Cartesian station coordinates for people using high quality orbits
    sp3 : bool
        whether you use multi-GNSS sp3 file to do azimuth elevation angle calculations
    gzip: bool
        gzip compress snrfiles. No idea if it is used here ...
        as this compression should happen in the calling function, not here
    orb : str
        requested orbit source
    hour : int
        ultrarapid orbit hour
    """
    # decimation
    idec = int(dec)
    missing = True
    station = fname.lower() ; station = station[0:4]
    yy,month,day, cyyyy, cdoy, YMD = g.ydoy2useful(year,doy)
    gfz_date = 2024 + 153/365.25 # when we added the new GFZ directory to gnssrefl
    
    if sp3:
        if orb is not None:
            if (orb == 'wum2'):
                xf,orbdir,foundit = g.get_wuhan_orbits(year,month,day,hour)
                if not foundit:
                    print('try noon previous day')
                    if doy == 1:
                        xf,orbdir,foundit = g.get_wuhan_orbits(year-1,12,31,12)
                    else:
                        xf,orbdir,foundit = g.get_wuhan_orbits(year,doy-1,0,12)

            if orb == 'ultra':
                # first try given doy
                xf,orbdir,foundit = g.new_ultra_gfz_orbits(year,doy,0,hour=0)
                # now try doy before
                if not foundit:
                    if (doy == 1):
                        xf,orbdir,foundit = g.new_ultra_gfz_orbits(year-1,12,31,hour)
                    else:
                        xf,orbdir,foundit = g.new_ultra_gfz_orbits(year,doy-1,0,hour)
                        if not foundit:
                            print('Try ultrarapid orbit from noon on the previous day')
                            if ((doy-1) == 1):
                                xf,orbdir,foundit = g.new_ultra_gfz_orbits(year-1,12,31,12)
                            else:
                                xf,orbdir,foundit = g.new_ultra_gfz_orbits(year,doy-2,0,12)
        else:
            # the old way is to try multiple orbit sources
            if (year+doy/365.25) < gfz_date:
                # someone needs to link to new code for newer final orbits
                xf,orbdir,foundit=g.gbm_orbits_direct(year,month,day)
            else:
                print('Try Rapid orbit products from GFZ ')
                xf,orbdir,foundit=g.rapid_gfz_orbits(year,month,day)

                if not foundit: 
                    print('Could not find the rapid orbits from GFZ. Try ultrarapid')
                    if (year + doy/365.25) > gfz_date:
                        print('Use new GFZ directory for ultra rapid orbits with the long filenames')
                        xf,orbdir,foundit = g.new_ultra_gfz_orbits(year,doy, 0,hour=0)
                        if not foundit:
                            if doy == 1:
                                xf,orbdir,foundit = g.new_ultra_gfz_orbits(year-1,12,31,hour=0)
                            else:
                                xf,orbdir,foundit = g.new_ultra_gfz_orbits(year,doy-1,0,hour=0)
                    else:
                        print('use old GFZ directory structure')
                        xf,orbdir,foundit = g.ultra_gfz_orbits(year,month,day,0)
                    if not foundit:
                        print('Could not find the ultrarapid orbits from GFZ. Trying Wuhan')
                        xf,orbdir,foundit = g.get_wuhan_orbits(year,month,day,hour)
                    if not foundit:
                        print('Out of luck - could not find a good orbit file for you')
                        return

        # define orbit file name
        orbfile = orbdir + '/' + xf # hopefully
        print('Found a sp3 file: ', orbfile)

    # this is to help a colleague 
    if station == 'argt':
        gt.decipher_argt(station,'jul01.txt',idec,snrfile,orbfile,recv,csnr,year,month,day)
        if os.path.isfile(snrfile):
            print('File made, ignoring the rest of the code')
            return
        else:
            print('Translation was unsuccessful'); return


    #check whether the input file is a uncompressed or compressed     
    if os.path.exists(locdir + fname):
        subprocess.call(['cp', '-f',locdir + fname ,'.'])
        t, prn, az, elv, snr, freq = read_nmea(fname)#read nmea files
        subprocess.call(['rm',fname])
        missing = False

    if os.path.exists(locdir + fname + '.gz') and missing:
        subprocess.call(['cp', '-f',locdir + fname + '.gz' ,'.'])
        subprocess.call(['gunzip', '-f', fname + '.gz'])
        t, prn, az, elv, snr, freq = read_nmea(fname)#read nmea files
        subprocess.call(['rm',fname])
        missing = False
        
    if os.path.exists(locdir + fname + '.Z') and missing:
        subprocess.call(['cp', '-f',locdir + fname + '.Z','.'])
        subprocess.call(['uncompress', fname + '.Z'])
        t, prn, az, elv, snr, freq = read_nmea(fname)#read nmea files
        subprocess.call(['rm',fname])
        missing = False
        
    print('Number of t values ', len(t))
    # why is there all this going back and forth between lists and np arrays?

    t = np.array(t);az = np.array(az);elv = np.array(elv);snr = np.array(snr);prn = np.array(prn); freq=np.array(freq)
    #remove empty records
    if (station != 'argt'):
        #????
        t = t[az !=''];snr = snr[az !=''];prn = prn[az !=''];elv = elv[az !=''];freq = freq[az != ''];az = az[az !=''];
        t = t[elv !=''];az = az[elv !=''];snr = snr[elv !=''];prn = prn[elv !=''];freq = freq[elv != ''];elv = elv[elv !='']
        t = t[snr !=''];az = az[snr !=''];prn = prn[snr !=''];elv = elv[snr !=''];freq = freq[snr != ''];snr = snr[snr !='']
        t = t[prn !=''];az = az[prn !=''];elv = elv[prn !=''];snr = snr[prn !=''];freq = freq[prn != ''];prn = prn[prn !=''] 
        t = t[freq !=''];az = az[freq !=''];elv = elv[freq !=''];snr = snr[freq !=''];prn = prn[freq !='']; freq = freq[freq != '']
    
    az = az.astype(float)
    elv = elv.astype(float)
    snr = snr.astype(float)
    prn = prn.astype(int)

    prn_unique = np.unique(prn) 
    #print(prn_unique)

    T = []; PRN = []; AZ = []; ELV = []; SNR = []; FREQ = []        
    for i_prn in prn_unique:
        # the original code added 100 - but did not take into account the 
        # satellite numbers have been shifted for glonass.
        # also there is an illegal signal at satellite "48" 
        # i do not know what that is, but i am ignoring it.
        #if (i_prn > 32):
        #    print(i_prn, 'looks like an illegal satellite number')
        time = t[prn == i_prn];angle = elv[prn == i_prn];azimuth = az[prn == i_prn]; frequency = freq[prn == i_prn]
        Snr = snr[prn == i_prn];Prn = prn[prn == i_prn]
        
        angle_fixed, azim_fixed = fix_angle_azimuth(time, angle, azimuth)#fix the angles 
        if (len(angle_fixed) == 0 and len(azim_fixed) == 0):
            continue
            
        T.extend(time);AZ.extend(azim_fixed);ELV.extend(angle_fixed);SNR.extend(Snr);PRN.extend(Prn); FREQ.extend(frequency)
        
        del  time, angle, azimuth, Snr, Prn, angle_fixed, azim_fixed, frequency
        
    # It is easier for the sp3 option to write out the time, satellite, and SNR data into a plain file.
    # then the fortran can read that file and calculate the orbits from teh SP3 file and write out a new 
    # file with the correct azimuth and elevation angle.
    leap_mjd = g.getMJD(year,month,day,0)

    offset = g.read_leapsecond_file(leap_mjd)
    print('Leap second offset ', offset)

    if sp3 :
       # this tmpfile has a terrible name.  
        #tmpfile =  station + 'tmp.txt'
        tmpfile = g.randomfilename() + '.txt'
        timetags = np.unique(T)

        xT = np.asarray(T)
        xPRN = np.asarray(PRN)
        xSNR = np.asarray(SNR)
        xfreq = np.asarray(FREQ)
        print('Opening temporary file : ', tmpfile)
        fout = open(tmpfile, 'w+')
        fout.write('{0:15.4f}{1:15.4f}{2:15.4f} \n'.format(recv[0], recv[1],recv[2]) )
        fout.write('{0:6.0f}{1:6.0f}{2:6.0f} \n'.format(year, month, day) )
        #(t[i], prn[i], snr[i],freq[i])
        # look thru the unique time tags
        print('Found ', len(timetags), ' time tags')
        for i in range(0,len(timetags)):
            if ( (timetags[i] % idec) == 0):
                # for this timetag and if pass decimation test
                jj = (xT == timetags[i])
                # satellites at the epoch
                unique_sats_epoch = np.unique(xPRN[jj])
                epoch_sat = xPRN[jj]
                # snr data at the epoch
                epoch_snr = xSNR[jj]
                # fr data at the epoch
                epoch_freq = xfreq[jj]
                # go thru the unique satellite numbers at this epoch... 
                for ij in range(0,len(unique_sats_epoch)):
                    # pick out that satellite number
                    sat = int(unique_sats_epoch[ij])
                    # find all data that go with this 
                    ijk = (sat == epoch_sat)
                    # store the SNR data ..
                    snrdata = epoch_snr[ijk]
                    # store the freq data ..
                    frdata = epoch_freq[ijk]
                    # to fix the long-lived glonass mistake
                    if (sat > 100) & (sat < 200):
                        sat = sat - 64

                    #print(timetags[i], sat, snrdata, frdata)
                    # now write them out ... first assume all SNR data are zero
                    # then reassign based on frequency
                    s1=0; s2=0; s5=0  
                    for iugh in range(0,len(frdata)):
                        if (frdata[iugh] == '1'):
                            s1 = snrdata[iugh]
                        elif (frdata[iugh] == '2'):
                            s2 = snrdata[iugh]
                        elif (frdata[iugh] == '5'):
                            s5 = snrdata[iugh]
                    #print(timetags[i], sat, s1, s2, s5)
                    # testing for leap seconds, not good for all time
                    fout.write('{0:8.0f} {1:3.0f} {2:6.2f} {3:6.2f} {4:6.2f} \n'.format(timetags[i]+offset, sat, s1, s2, s5) )
                    
        fout.close()
        print(station, tmpfile,snrfile)
        print(orbfile)
        gt.new_azel(station,tmpfile,snrfile,orbfile,csnr)
        print('Az/El Updated...')
        return # translation has taken place in new_azel, so return to main code 

    # this was my first effort.  It only allowed L1 data. I am not deleting it - but
    # it is no longer valid.
    if sp3 and False:
        tmpfile =  station + 'tmp.txt'
        print('Opening temporary file : ', tmpfile)
        fout = open(tmpfile, 'w+')
        fout.write('{0:15.4f}{1:15.4f}{2:15.4f} \n'.format(recv[0], recv[1],recv[2]) )
        fout.write('{0:6.0f}{1:6.0f}{2:6.0f} \n'.format(year, month, day) )
        #(t[i], prn[i], snr[i],freq[i])
        for i in range(0,len(T)):
            if ( (int(t[i]) % idec) == 0):
                sat = PRN[i]
                # glonass satellites are misnamed by the code
                if (sat > 100) & (sat < 200):
                    sat = sat - 64
                fout.write('{0:8.0f} {1:3.0f} {2:6.2f} {3:s} \n'.format(T[i], sat, SNR[i], freq[i]) )
                newl = [T[i], sat, SNR[i], int(freq[i])]
        fout.close()
        #subprocess.call(['cp', tmpfile,'k_debug.txt']) 
        # make the snrfile
        gt.new_azel(station,tmpfile,snrfile,orbfile,csnr)
        print('Az/El Updated...')

        return # - in theory the fortran called in new_azel took care of everything
    
    inx = np.argsort(T)  #Sort data by time
    
    T = np.array(T);PRN = np.array(PRN);ELV = np.array(ELV);SNR = np.array(SNR);AZ = np.array(AZ); FREQ=np.array(FREQ)
    
    T = T[inx];PRN = PRN[inx];ELV = ELV[inx];SNR = SNR[inx];AZ = AZ[inx]; FREQ=FREQ[inx]
                               
    emin,emax = elev_limits(int(csnr))#select snr option 50, 66, 88, 99
    #write to an output file 
    with open(snrfile, 'w') as fout:
        for i in range(len(T)):
            if (float(ELV[i]) >= emin) and (float(ELV[i]) <= emax):
                
                #Via: https://receiverhelp.trimble.com/alloy-gnss/en-us/NMEA-0183messages_GSV.html
                
                #Note - Not sure if this is the best way to do this, or rather do it below when we can decide what
                #constellation we're looking at from the GxGSV line, and then add/subtract from the PRN there?
                #That might make it easier to keep straight
                
                #GPS
                if PRN[i] < 100:
                    p = float(PRN[i])
                    #if p > 32:
                        #$GPGSV indicates GPS and SBAS satellites. If the PRN is greater than 32, 
                        #this indicates an SBAS PRN, 87 should be added to the GSV PRN number to determine the 
                        #SBAS PRN number.
                        ##p += 87 ### Not sure how this could interact with future processing steps...
                
                #GLONASS
                elif (PRN[i] > 100 and PRN[i] < 200):
                    #$GLGSV indicates GLONASS satellites. 64 should be subtracted from the GSV PRN number 
                    #to determine the GLONASS PRN number.
                    p = float(PRN[i]) - 64
                
                #GALILEO
                elif (PRN[i] > 200 and PRN[i] < 300): 
                    #$GAGSV indicates Galileo satellites.
                    p = float(PRN[i])
                    
                #BEIDOU    
                elif (PRN[i] > 300 and PRN[i] < 400): 
                    #$GBGSV indicates BeiDou satellites. 100 should be subtracted from the 
                    #GSV PRN number to determine the BeiDou PRN number.
                    p = float(PRN[i]) # -100 ###Not sure how this could interact with future processing steps...
    
                #Final output format (https://github.com/kristinemlarson/gnssrefl/blob/master/docs/pages/rinex2snr.md)
                #Satellite number (remember 100 is added for Glonass, etc)
                #Elevation angle, degrees
                #Azimuth angle, degrees
                #Seconds of the day, GPS time
                #elevation angle rate of change, degrees/sec.
                #S6 SNR on L6
                #S1 SNR on L1
                #S2 SNR on L2
                #S5 SNR on L5
                #S7 SNR on L7
                #S8 SNR on L8        
                
                # it is not necessary to write out l7 and l8.  or l6. zero can be put there.
                l1 = 0; l2 = 0; l5 = 0; l6 = 0; l7 = 0; l8 = 0
                
                #Check what frequency we are at to know where to write each line
                f_store = FREQ[i]
                # this will create extremely large files ...
                #print(t[i], f_store, SNR[i])
                if f_store == '1':
                    l1 = float(SNR[i])
                elif f_store == '2':
                    l2 = float(SNR[i])
                elif f_store == '5':
                    l5 = float(SNR[i])
                # remove l6 and l7 ... we can add back in if a cheap instrument ever produces these obs
                #elif f_store == '6':
                #    l6 = float(SNR[i])
                #elif f_store == '8':
                #    l8 = float(SNR[i])
                    
                #NOTE -- All satellites have a 'mixed/undefined' frequency 0 for 
                # NMEA 4.11 (https://gpsd.gitlab.io/gpsd/NMEA.html#_nmea_4_11_system_id_and_signal_id)
                #This is never used here so could leave blank snr lines!
                
                outline = "%3g %10.4f %10.4f %10g %4s " % (p, float(ELV[i]), float(AZ[i]), float(T[i]), '0')
                snrline = "%7.2f %7.2f %7.2f %7.2f " % (l6, l1, l2, l5 )
                 
                # apply decimating here
                if ( (int(T[i]) % idec) == 0):
                    fout.write(outline + snrline + '\n')
                #fout.write("%3g %10.4f %10.4f %10g %4s %4s %7.2f %4s %4s\n" % (p, float(ELV[i]), float(AZ[i]), float(T[i]),'0', '0', float(SNR[i]),'0', '0')) 
        
def read_nmea(fname):
    """
    reads a NMEA file.

    KL: is this statement correct?
    "it only reads the GPGGA sentence (includes snr data) in NMEA files"

    Parameters
    ----------
    fname : str
        NMEA filename

    Returns
    -------
    t : list of int
        timetags in GPS seconds 

    prn : list of int
        GPS satellite numbers

    az : list of floats ??
        azimuth values (degrees)

    elv : list of floats ??
        elevation angles (degrees)

    snr : list of floats
        snr values

    freq : list of ???
        apparently frequency values - 

    """
    
    ### SOME USEFUL LINKS
    #https://cddis.nasa.gov/sp3c_satlist.html
    #https://receiverhelp.trimble.com/alloy-gnss/en-us/NMEA-0183messages_MessageOverview.html
    #https://receiverhelp.trimble.com/alloy-gnss/en-us/NMEA-0183messages_GSV.html
    
    #Could also go directly to reading the file via gzip.open(fname, 'rb') or gzip.open(fname, 'rt') for text
    with open(fname, 'rb') as f:
        lines = f.readlines()

    t = []; prn = []; az = []; elv = []; snr = []; freq = []
    
    curdt = None #Empty starting date to ensure only get data with proper timing
    t_sec = None #Empty starting time
    for i, line in enumerate(lines):
        line = line.decode('utf-8')
        
        if '$' in line[1:-1]: #Skip line if misplaced $ via https://github.com/purnelldj/gnssr_lowcost
            continue
        if not line.startswith('$G'): #Skip line if improper start
            continue
        if len(line.split('*')) > 2: #Additional error checking for misplaced * or combined lines
            continue
        
        #line = line.replace('$', '') #Strip leading $
        
        if 'RMC' in line: #Get timing info
            row = line.split(',')
            #sect = int(float(row[1][0:2]) * 60 * 60 + float(row[1][2:4]) * 60 + float(row[1][4:])) #Time in seconds
            try:
                curdt = datetime.datetime(int(row[9][4:6])+2000,int(row[9][2:4]),int(row[9][0:2])) #Current date
            except:
                pass
        
        if not curdt: #Skip forward until the first 'RMC' instance which gives a proper date to store
            continue
        
        if 'GGA' in line: #read GPGGA sentence: Global Positioning System Fix Data 
            try:
                hr = int(line.split(",")[1][0:2])
                mn = int(line.split(",")[1][2:4])
                sc = float(line.split(",")[1][4:8])
                t_sec = hr*3600 + mn*60 + sc
                if (i > 100 and t_sec == 0):                   #set t to 86400 for the midnight data
                    t_sec = 86400
            except:
                t_sec = None
                continue # skip invalid GGA sentence
            #This could be added if positioning info is interesting/needed. Via https://github.com/purnelldj/gnssr_lowcost
            #   row = line.split(',')
            #   if row[3] == 'S': #Northern or southern hemisphere
            #        latt = -latt
            #   lont = float(row[4][0:3]) + float(row[4][3:])/60 #Longitude
            #   if row[5] == 'W':
            #        lont = -lont #Eastern or Western hemisphere
            #   hgtt = float(row[9])+float(row[11]) #Orthometric height (MSL ref)

        #Read Satellite positions (GSV) - The GSV message string identifies the number of SVs in view, the PRN numbers, 
        #elevations, azimuths, and SNR values.
        elif 'GSV' in line:                    #read GPGSV sentence: GPS Satellites in view in this cycle   
            if t_sec is None: # Discard line if there's no valid timestamp
                continue
            try:
                sent = line.split(",")         #GSV sentence 
                #ttl_ms = int(sent[1])          #Total number of messages in the GPGSV sentence 
                #ms = int(sent[2])              #Message number 
                
                #Check for a proper length GSV line (avoid mixed/incomplete lines)
                if len(sent) not in [21, 20, 17, 16, 13, 12, 9, 8]:
                    continue
            
                if "GPGSV" in line: #GPS
                    prn_offset = 0
                elif "GLGSV" in line: #Glonass
                    prn_offset = 100
                elif "GAGSV" in line: #Galileo
                    prn_offset = 200
                elif "BDGSV" in line: #Beidou
                    prn_offset = 300
                elif 'GBGSV' in line: #Also Beidou, depending on chip
                    prn_offset = 300
                else:
                    print('Undefined constellation')
                    continue
                
                #Note: $GNGSV uses PRN in field 4. Other $GxGSV use the satellite ID in field 4. 
                    #Jackson Labs, Quectel, Telit, and others get this wrong, in various conflicting ways.
                #GNGSV is not currently supported in this code!
                
                if len(sent) in [21, 17, 13, 9]: #Some chips spit out an extra piece of info at the end 
                    #Ublox-9 etc has a signal ID which can be used to split between L1, L2, etc
                    #    #https://gpsd.gitlab.io/gpsd/NMEA.html#_nmea_4_11_system_id_and_signal_id
                    #This seems to be a NMEA 4.11 feature -- not all chips will have this!
                    
                    sig_id = sent[-1].split('*')[0] #Store that signal ID to check frequency
                    sent = sent[:-1] #Remove the last line for subsequent SNR data gatheriing
                    if "GPGSV" in line:
                        if sig_id in ['0','1','2','3']:
                            sig_id = 1
                        elif sig_id in ['5', '6']:
                            sig_id = 2
                        elif sig_id in ['7', '8']:
                            sig_id = 5
                    if 'GLGSV' in line:
                        if sig_id in ['0', '1', '2']:
                            sig_id = 1
                        elif sig_id in ['3', '4']:
                            sig_id = 2
                    if 'GAGSV' in line:
                        if sig_id in ['6', '7']:
                            sig_id = 1
                        elif sig_id in ['1', '2', '3']:
                            sig_id = 5
                        elif sig_id in ['4', '5']:
                            sig_id = 6
                    if ('GBGSV' in line or 'BDGSV' in line): #Beidou has two possible calls in NMEA (and weird frequency codes)...
                        if sig_id in ['0','1','2','3','4']:
                            sig_id = 1
                        elif sig_id in ['5','6','7','B','C']:
                            sig_id = 2
                        elif sig_id in ['8','9','A']:
                            sig_id = 3
            
                else:
                    sig_id = 1 #Default is L1 -- this is most data/chips that return NMEA by default
                    
                for inds in [4, 8, 12, 16]: #Check for a max of four satellites in view per line
                    try:
                        sat_prn = int(sent[inds]) + prn_offset #field 4,8,12,16 :  SV PRN number
                        elev = sent[inds + 1] #field 5,9,13,17 :  Elevation in degrees, 90 maximum
                        azi = sent[inds + 2] #field 6,10,14,18:  Azimuth in degrees
                        snr_ = sent[inds + 3] #field 7,11,15,19:  SNR, 00-99 dB (null when not tracking)
                        
                        #Note: NMEA 4.1+ systems (u-blox 9, Quectel LCD79) may emit an extra field, Signal ID, 
                            #just before the checksum -- via: https://gpsd.gitlab.io/gpsd/NMEA.html#_gsv_satellites_in_view
                        #Hence .split('*')[0]
                        snr_ = snr_.split('*')[0]
                        
                        prn.append(str(sat_prn))
                        elv.append(elev)
                        az.append(azi)
                        snr.append(snr_)
                        freq.append(str(sig_id))
                        t.append(t_sec)
                        
                    except: #Error handle for poorly captured data (e.g., no PRN, or at max nr of satellites in line)
                        pass
                    
            except:
                #Sometimes you might have a mixed line due to sensors turning on/off unexpectedly
                print(line, 'Failed! Skipping...')
                #traceback.print_exc()
    return t, prn, az, elv, snr, freq

def fix_angle_azimuth(time, angle, azimuth):
    """
    interpolate elevation angles and azimuth to retrieve decimal values thru time     
    this is for NMEA files. This is not used if sp3 orbit file used.

    Parameters
    ----------
    time : list of floats
        GPS seconds of the week
    angle : list of floats
        elevation angles  (degrees)

    azimuth : list of floats
        azimuth angles (degrees)

    Returns
    -------
    angle_fixed : list of floats
        interpolated elevation angles

    azim_fixed : list of floats
        interpolated azimuth angles

    """
    
#delet nans
    time = time[~np.isnan(angle)];azimuth =  azimuth[~np.isnan(angle)];angle = angle[~np.isnan(angle)]
#delet nans
    time = time[~np.isnan(azimuth)];angle = angle[~np.isnan(azimuth)];azimuth =  azimuth[~np.isnan(azimuth)]
#delet nans
    angle = angle[~np.isnan(time)];azimuth =  azimuth[~np.isnan(time)];time = time[~np.isnan(time)]

    dangle = np.diff(angle)#diff angle
    ind1 = np.array(np.where(dangle!=0))
    
    dazim = azimuth_diff1(azimuth)#diff azimuth
    ind3 = np.array(np.where(dazim!=0))
    
    if ind1.size > 1 and ind3.size > 1:
        ind2 = ind1 +1
        time_angle0 = ( (time[ind1]+time[ind2])/2.0 )[0]
        angle0 = ( (angle[ind1]+angle[ind2])/2.0 )[0]

        ind4 = ind3 +1
        time_azim0 = ( (time[ind3]+time[ind4])/2.0 )[0]
        azim0 = azimuth_mean(azimuth[ind3], azimuth[ind4])

#interpolate at mean values
        f_ang = interp1d(time_angle0, angle0, kind = 'linear', fill_value="extrapolate")
        f_azim = interp1d(time_azim0, azim0, kind = 'linear', fill_value="extrapolate")
        angle_fixed = f_ang(time)#interpolated elev angle
        azim_fixed = f_azim(time)#interpolated azimuth angle

    else:#return emty outputs for tracks with not changing elv and/or azm values. This often happnes for a chunk of nmea data when the daily record is not complete  
        angle_fixed = []; azim_fixed = []
    
    return angle_fixed, azim_fixed  
        
def azimuth_diff2(azim1, azim2):
    diff = azim1 - azim2
    idx = (diff > +180)  
    diff[idx] = diff[idx] - 360
    idx = (diff < -180)  
    diff[idx] = diff[idx] + 360
    return diff

def azimuth_diff1 (azim):
    """
    Parameters
    ----------
    azim: ??

    """
    azim_a = azim[0:-1]
    azim_b = azim[1:]
    diff = azimuth_diff2 (azim_b, azim_a);
    #diff = azimuth_diff2 (azima, azimb);  % WRONG!   
    return diff

def azimuth_diff(azim1, azim2):
    """
    someone should document this

    Parameters
    ----------
    azim1 : ??

    azim2 : ??

    Returns
    -------
    ???

    """
    if not(azim2.size):
        diff = azimuth_diff1 (azim1)
    else:
        diff = azimuth_diff2 (azim1, azim2)
#    diff = np.abs(diff)
    return diff
     
def angle_range_positive(ang):
    """
    someone should document this

    Parameters
    ----------
    ang : ??

    """
#    idx1 = np.isfinite(ang)
    ang = np.angle(np.exp(1j*ang*np.pi/180))*180/np.pi  
    idx2 = (ang < 0)
    ang[idx2] = 360 + ang[idx2]
    return ang

def azimuth_mean(azim1, azim2):
    """
    someone that is not me should document this

    Parameters
    ----------
    azim1 : list of floats ? 
         azimuth degrees

    azim2 : list of floats
         azimuth degrees

    Returns
    -------
    azim : list of floats ?
        azimuths in degrees
    """
    azim = np.concatenate([azim1, azim2])
    if np.all(azim1 >= 0) and np.all(azim2 >= 0):
        azim1 = angle_range_positive(azim1)
        azim2 = angle_range_positive(azim2)
        azim = ( (azim1 + azim2)/2.0 )[0] 
    else:
        x1 = np.sin(azim1*np.pi/180);x2 = np.sin(azim2*np.pi/180)
        y1 = np.cos(azim1*np.pi/180);y2 = np.cos(azim2*np.pi/180) 
        x = ( (x1 + x2)/2.0 )[0];y = ( (y1 + y2)/2.0 )[0]
        azim = 180/np.pi * np.arctan2(x, y)

    return azim

def quickname(station,year,cyy, cdoy, csnr):
    """
    Creates a full name of the snr file name (i.e. including the path) 
    >>>> Checks that directories exist.

    Parameters
    ----------
    station : str
        4 ch station name

    year : int
        full year

    cyy : str 
        two character year

    cdoy : str
        three character day of year

    csnr : str
        snr type, e.g. '66' 

    Returns
    -------
    fname : str
        output filename

    """
    
    xdir  = os.environ['REFL_CODE'] + '/'
    cyyyy = str(year)
    fname =  xdir + cyyyy + '/snr/' + station + '/' + station + cdoy + '0.' + cyy + '.snr' + csnr
    d = xdir + cyyyy
    if not os.path.isdir(d):
        subprocess.call(['mkdir', d])
    d = xdir + cyyyy + '/snr'
    if not os.path.isdir(d):
        subprocess.call(['mkdir', d])
    d = xdir + cyyyy + '/snr/' + station
    if not os.path.isdir(d):
        subprocess.call(['mkdir', d])

    return fname

def elev_limits(snroption):
    """

    Given a snr option, return the elevation angle limits

    Parameters
    ----------
    snroption : int
        snrfile number

    Returns
    -------
    emin : float
        min elevation angle (degrees)
    emax : float
        max elevation angle (degrees)

    """

    if (snroption == 99):
        emin = 5; emax = 30
    elif (snroption == 50):
        emin = 0; emax = 10
    elif (snroption == 66):
        emin = 0; emax = 30
    elif (snroption == 88):
        emin = 0; emax = 90
    else:
        emin = 5; emax = 30

    return emin, emax
  
def run_nmea2snr(station, year, doy, isnr, overwrite, dec, llh, recv, sp3, gzip,orb,hour):
    """
    runs the nmea2snr conversion code - ONE DAY AT A TIME (2024 March 16)

    Looks for NMEA files in $REFL_CODE/nmea/ssss/2023 for station ssss and year 2023.
    I prefer lowercase station names, but I believe the code allows both upper and lower
    case. 

    Files are named:  SSSS1520.23.A 

    where SSSS is station name, day of year 152 and 
    the last two characters of the 2023 as the middle value.

    The SNR files are stored with upper case if given upper case, lower case if given lower case.

    Parameters
    ----------
    station : str
        4 ch name of station 
    year : int
        full year 
    doy : int
        day of year
    isnr : int
        snr file type
    overwrite : bool
        whether make a new SNR file even if one already exists
    dec : int
        decimation in seconds
    llh : list of floats
        lat and lon (deg) and ellipsoidal ht (m)
    recv : list of floats
        cartesian receiver coordinates (m)
    sp3 : bool
        whether you want to use GFZ rapid sp3 file for the orbits
    gzip : bool
        whether snrfiles are gzipped after creation
    orb : str
        requested orbit source
    hour : int
        requested hour for ultrarapid orbit

    """
    # avoiding makan's loops - but using True to avoid re-indenting
    if True:
        locdir= os.environ['REFL_CODE'] + '/nmea/' + station + '/' + str(year) + '/'
        if True:
            yy,mm,dd, cyyyy, cdoy, YMD = g.ydoy2useful(year,doy)
            csnr = str(isnr) ; cyy =  cyyyy[2:4]
            snrfile =  quickname(station,year,cyy,cdoy,csnr)#snr filename
            snre = g.snr_exist(station,year,doy,csnr)#check if snrfile already sxists
            if snre:
                if overwrite:
                    print('SNR file exists, but you requested it be overwritten')
                    # just in case you have a previously gunzipped version
                    if os.path.exists(snrfile):
                        subprocess.call(['rm', snrfile])
                        snre = False
                    if os.path.exists(snrfile + '.gz'):
                        subprocess.call(['rm', snrfile + '.gz'])
                        snre = False
                else:
                    print('SNR file already exists', snrfile)
        
            illegal_day = False
            if ( doy > g.dec31(year) ):
                illegal_day = True
        
            if (not illegal_day) and (not snre):
                r =  station + cdoy + '0.' + cyy + '.A'# nmea file name example:  WESL2120.21.A 
                if os.path.exists(locdir+r) or os.path.exists(locdir+r+'.gz') or os.path.exists(locdir+r+'.Z') or (station == 'argt'):
                    nmea_translate(locdir, r, snrfile, csnr, dec, year, doy, recv, sp3, gzip,orb,hour)
                    if os.path.isfile(snrfile):
                        print('SUCCESS: SNR file created', snrfile)
                    if os.path.isfile(locdir + r ):
                        # gzip the NMEA file now
                        print('gzip the NMEA file', locdir + r)
                        subprocess.call(['gzip', locdir + r])
                        # otherwise it is already gzipped?
                    if gzip:
                        if not snrfile.endswith('.gz'):
                            subprocess.call(['gzip', snrfile])
                            print('SNR file gzip compressed')
                else:
                    print('NMEA file '+ locdir + r +' does not exist')



# -*- coding: utf-8 -*-
import argparse
import numpy as np
import os
import subprocess
import sys

import gnssrefl.gps as g

def main():
    """
    decimates a SNR file

    Examples
    --------

    smoosh_snr p041 2015 175 5
        decimates an existing SNR file (year 2015 and day of year 175) to 5 seconds


    Parameters
    ----------
    station : str
        four ch station name
    year : int
        full year
    doy : int
        day of year 
    dec : int
        decimation value in seconds
    snr : int, optional
        snr file name, i.e. 99 or 10, default is 66
    doy_end : int, optional
        allows you to analyze data from doy to doy_end

    """

    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="4 ch station name", type=str)
    parser.add_argument("year", help="year", type=int)
    parser.add_argument("doy", help="day of year", type=int)
    parser.add_argument("dec", help="decimate value, sec", type=int)
    parser.add_argument("-snr", help="snr value", type=int,default=None)
    parser.add_argument("-doy_end", help="decimation value, seconds", type=int,default=None)

    args = parser.parse_args()
    station = args.station
    year = args.year
    doy = args.doy
    dec = args.dec

    if (args.snr is None):
        snr = 66
    else:
        snr = args.snr

    csnr = str(snr)

    cyyyy, cyy, cdoy = g.ydoych(year,doy)
    myxdir = os.environ['REFL_CODE'] + '/' +  cyyyy +  '/snr/' + station + '/'
    if args.doy_end is None:
        doy_end = doy
    else:
        doy_end = args.doy_end

    for d in range(doy, doy_end+1):
        cyyyy, cyy, cdoy = g.ydoych(year,d)
        snrfile = myxdir +  station + cdoy + '0.' + cyy + '.snr' + csnr
        snrfilet = snrfile + 't'

        if not os.path.exists(snrfile):
            print('snrfile does not exist: ', snrfile)
        else:
            a=np.loadtxt(snrfile)
            i = (a[:,3] % dec) == 0 ; a=a[i,:]
            nr,nc = a.shape
            fout = open(snrfilet, 'w+')
            if (nc == 11):
                np.savetxt(fout, a, fmt="%3.0f %9.4f %9.4f %9.1f %11.6f %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f")
            elif (nc == 9):
                np.savetxt(fout, a, fmt="%3.0f %9.4f %9.4f %9.1f %11.6f %6.2f %6.2f %6.2f %6.2f")
            else:
                print('Incorrect number of columns. Exiting without making new file')
                sys.exit()
            fout.close()
            print('new/decimated snrfile created')
            subprocess.call(['mv','-f',snrfilet,snrfile])

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import argparse
import datetime
import numpy as np
import os
import requests
import sys
import matplotlib.pyplot as plt
import subprocess
import urllib

import gnssrefl.gps as g

#from gnssrefl.utils import validate_input_datatypes, str2bool



def download_psmsl(station: str, output: str = None, plt: bool = False):
    """
    Downloads PSMSL tide gauge files created by Simon Williams 
    in json format, converts it to plain txt or csv format

    Parameters
    ----------

    station : str
        4 ch station name

    output : str, optional
        Optional output filename
        default is None

    plt: bool, optional
        plot comes to the screen

    """
    g.check_environ_variables()
    xdir = os.environ['REFL_CODE']
    outdir = xdir  + '/Files/'
    if not os.path.exists(outdir) :
        subprocess.call(['mkdir', outdir])


    csv = False
    if output is None:
    # use the default
        outfile = outdir + station + '_' + 'psmsl.txt'
    else:
        outfile = outdir + output
        if output[-3:] == 'csv':
            csv = True

    filename = station + '.zip'
    cfilename = station + '.csv'
    url = 'https://psmsl.org/data/gnssir/data/main/' + filename
    print(url)

    # remove old files 
    if os.path.exists(filename):
        subprocess.call(['rm', '-f',filename])
    if os.path.exists(cfilename):
        subprocess.call(['rm', '-f',cfilename])

    print('Pick up the files')

    try:
        urllib.request.urlretrieve(url, filename)
        subprocess.call(['unzip', filename])
        subprocess.call(['rm', '-f', filename])
        # this should produce a csv file - cfilename
    except:
        print('Some problem with the download. Perhaps the station does not exist')
        sys.exit()

    if os.path.exists(cfilename):
       # read the file
        obs, simon_mjdish, sl,sprn,sfr, simon_az = g.read_simon_williams(cfilename,outfile)
    else:
        print('some problem with the download')
        sys.exit()
    

    NV = len(obs)

    if NV < 1:
        print('No data found')
        sys.exit()

    if plt:
        g.quickp('PSMSL ' + station,obs,sl)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import argparse
import datetime
import numpy as np
import os
import requests
import subprocess
import sys
import wget
import gnssrefl.gps as g

#from gnssrefl.utils import validate_input_datatypes, str2bool

def download_noaa(station: str, date1: str, date2: str, output: str = None, plt: bool = False, datum: str = 'mllw', subdir: str = None):
    """
    Downloads NOAA tide gauge files and stores locally
    If you ask for 31 days of data or less, it will download exactly what you ask for.
    But if you want a longer time series, this code needs to query the NOAA API every month.
    To make the code easier to write, I start with the first day of the first month you ask for and end with
    last day in the last month.

    Output is written to REFL_CODE/Files/ unless subdir optional input is set
    Plot is sent to the screen if requested.

    Parameters 
    ----------
    station : str
        7 character ID of the station.

    date1 : str
        start date.
        Example value: 20150101

    date2 : str
        end date.
        Example value: 20150110

    output : string, optional
        Optional output filename
        default is None

    plt: boolean, optional
        plot comes to the screen
        default is None

    datum: string, optional
        set to lwd for lakes?
        default is mllw

    subdir : str, optional
        subdirectory for output in the $REFL_CODE/Files area

    """
    g.check_environ_variables()

    xdir = os.environ['REFL_CODE']
    outdir = xdir  + '/Files/'
    if not os.path.exists(outdir) :
        subprocess.call(['mkdir', outdir])

    if (subdir is None):
        print('Using regular output directory',outdir )
    else:
        outdir = xdir  + '/Files/' + subdir + '/'
        print('User requested a subdirectory for output', outdir)
        if not os.path.exists(outdir):
            subprocess.call(['mkdir', outdir])

    if len(station) != 7:
        print('station must have 7 characters ', station); sys.exit()
    if len(date1) != 8:
        print('date1 must have 8 characters', date1); sys.exit()
    if len(date2) != 8:
        print('date2 must have 8 characters', date2); sys.exit()


    csv = False
    # use the default, which is plain text
    if output is None:
        outfile = outdir + station + '_' + 'noaa.txt'
    else:
        if output[-3:] == 'csv':
            csv = True
        outfile = outdir + output

    print('Writing contents to: ', outfile)
    fout = open(outfile, 'w+')
    if csv:
        fout.write("{0:s} {1:s} \n".format('#', 'NOAA Station: ' + station ))
        fout.write("#YYYY,MM,DD,HH,MM, SS, Water(m),DOY,   MJD \n")
        #fout.write("#YYYY,MM,DD,HH,MM, Water(m),DOY,   MJD, Seconds\n")
    else:
        #fout.write("%YYYY MM DD HH MM  Water(m) DOY    MJD      Seconds\n")
        #fout.write("% 1    2  3  4  5    6       7       8         9\n")
        fout.write("{0:s} \n".format('%' + ' NOAA Station: ' + station ))
        fout.write("%YYYY MM DD HH MM  SS Water(m) DOY    MJD  \n")
        fout.write("% 1    2  3  4  5   6   7       8     9\n")


    year1, month1, day1, doy1,modjul1 = noaa2me(date1)
    year2, month2, day2, doy2,modjul2 = noaa2me(date2)
    deltad = int(modjul2-modjul1)
    if deltad < 1:
        print('second time is later than first time OR time is less than one day.')
        sys.exit()

    metadata = True
    if (deltad) > 31:
        tt,obstimes,slevel = multimonthdownload(station,datum,fout,year1,year2,month1,month2,csv)
        noaa_name = station # this is not hte right one ... but 
    else:
        tt = []; slevel = []; obstimes = []
        # 'data' are stored in the dictionary data
        data,error = pickup_from_noaa(station,date1,date2,datum,True)
        if not error:
            noaa_name = data['metadata']['name']
            print('WARNING: I have changed the format for this file.')
            tt,obstimes,slevel = write_out_data(data,fout, tt,obstimes,slevel,csv)

    fout.close()
    if plt:
        g.quickp(station,obstimes,slevel)

def noaa_command(station,fout,year,month1,month2,datum,metadata,tt,obstimes,slevel,csv):
    """
    downloads/writes NOAA tidegauge data for one month

    Parameters
    ----------
    station : str
        station name

    year : int
        full year

    month1 : int
        starting month

    month2 : int
        ending month

    datum : str
        water datum

    metadata : bool
        whether you want the metadata printed to the screen

    tt : numpy array 
        modified julian date for water measurements

    obstimes : numpy array of datetimes
        time of the measurements

    slevel : numpy array of floats
        water level in meters

    csv : bool
        True if csv output wanted (default is False)

    Returns
    -------
    tt : numpy array 
        modified julian date for water measurements

    obstimes : numpy array
        datetime format, updated with new data

    slevel : numpy array
        sea level (m) updated with new data

    """
    cyyyy = str(year)
    imeta = 0
    for m in range(month1,month2+1):
        cmm = '{:02d}'.format(m)
        d1 = cyyyy + cmm + '01'
        if m in [1,3,5,7,8,10,12]:
            d2 = cyyyy + cmm + '31'
        else:
            if m == 2:
                if (g.dec31(year) == 366): # leap year
                    d2 = cyyyy + cmm + '29'
                else:
                    d2 = cyyyy + cmm + '28'
            else:
                d2 = cyyyy + cmm + '30'

        print(d1,d2) # send whether imeta is equal to 0 ...  
        data, error = pickup_from_noaa(station,d1,d2,datum,imeta==0)
        if not error: # only write out if error not thrown. update the variables
            tt,obstimes,slevel = write_out_data(data,fout, tt,obstimes,slevel,csv)

        imeta = imeta + 1

    return tt,obstimes,slevel


def multimonthdownload(station,datum,fout,year1,year2,month1,month2,csv):
    """
    downloads NOAA water level measurements > one month

    Parameters
    ----------
    station : str
        NOAA station name

    datum : str
        definition of water level datum

    fout - fileID for writing results

    year1 : int
        year when first measurements will be downloaded

    month1 : integer
        month when first measurements will be downloaded

    year2 : integer
        last year when measurements will be downloaded

    month2 : integer 
        last month when measurements will be downloaded

    csv : boolean 
        whether output file is csv format

    Returns
    -------
    tt : list of times 
        modified julian day
    obstimes : list of datetime objects 

    slevel : list or is it numpy ?? 
         water level in meters

    """
    # set variables for making the plot ... 
    tt = []; slevel = []; obstimes = []
    metadata = True  # set true for now

    imeta = 0
    if (year1 == year2):
        tt, obstimes, slevel = noaa_command(station,fout,year1,month1,month2,datum,metadata,tt,obstimes,slevel,csv)
    else:
        for y in range(year1,year2+1):
            if (y == year1):
                tt,obstimes,slevel = noaa_command(station,fout,y,month1,12,datum,metadata,tt,obstimes,slevel,csv)
            elif (y == year2):
                tt,obstimes,slevel = noaa_command(station,fout,y,1,month2,datum,metadata,tt,obstimes,slevel,csv)
            else:  # get whole year
                tt,obstimes,slevel = noaa_command(station,fout,y,1,12,datum,metadata,tt,obstimes,slevel,csv)

    return tt, obstimes, slevel

def noaa2me(date1):
    """
    converts NOAA type of date string to simple integers

    Parameters 
    ----------
    date1 : string
        time in format YYYYMMDD for year month and day

    Returns 
    -------
    year1 : integer
        full year

    month1 : integer
        month

    day1 : integer
        day of the month

    doy : integer  
        day of year

    modjulday : float
        modified julian date

    """
    year1 = int(date1[0:4]); 
    month1=int(date1[4:6]); 
    day1=int(date1[6:8])
    doy, cdoy, cyyyy, cyy =g.ymd2doy(year1, month1, day1)

    mj, fj = g.mjd(year1, month1, day1, 0, 0, 0)
    modjulday = mj + fj;

    return year1, month1, day1, doy, modjulday

def write_out_data(data,fout, tt,obstimes,slevel,csv):
    """
    writes out the NOAA water level data to a file 
    20213-mar-27 using new format

    Parameters
    ----------
    data : dictionary from NOAA API

    fout : file ID 
        for output

    tt : ??

    obstimes : list of datetimes
        times of water level measurements

    slevel : numpy array of floats
        water level in meters

    csv : boolean
        whether csv format or not 

    Returns
    -------
    tt :  same as input, but larger

    obstimes : list of datetimes
        times for waterlevels

    slevel : list of floats
        water levels in meters

    """
    NV = len(data['data'])

    for i in range(0, NV):
        t = data['data'][i]['t']
        slr = data['data'][i]['v']
        slf = data['data'][i]['f']
        if slr == '':
            aa=1
            #print('no data')
        else:
            sl = float(data['data'][i]['v'])
            year = int(t[0:4]); mm = int(t[5:7]); dd = int(t[8:10])
            hh = int(t[11:13]); minutes = int(t[14:16])
            # seconds are zero
            ss = 0
            today = datetime.datetime(year, mm, dd)
            doy = (today - datetime.datetime(today.year, 1, 1)).days + 1
            m, f = g.mjd(year, mm, dd, hh, minutes, ss)
            mjd = m + f;
            tt.append(mjd)
            bigT = datetime.datetime(year=year, month=mm, day=dd, hour=hh, minute=minutes, second=ss)
            obstimes.append(bigT)

            slevel.append(sl)
            if csv:
                fout.write(" {0:4.0f},{1:2.0f},{2:2.0f},{3:2.0f},{4:2.0f},{5:2.0f},{6:7.3f},{7:3.0f},{8:15.6f} \n".format(year, mm, dd, hh, minutes, ss, sl, doy, mjd))
            else:
                fout.write(" {0:4.0f} {1:2.0f} {2:2.0f} {3:2.0f} {4:2.0f} {5:2.0f} {6:7.3f} {7:3.0f} {8:15.6f} \n".format(year, mm, dd, hh, minutes, ss, sl, doy, mjd))
            #if csv:
            #    fout.write(" {0:4.0f},{1:2.0f},{2:2.0f},{3:2.0f},{4:2.0f},{5:7.3f},{6:3.0f},{7:15.6f},{8:2.0f} \n".format(year, mm, dd, hh, minutes, sl, doy, mjd,ss))
            #else:
            #    fout.write(" {0:4.0f} {1:2.0f} {2:2.0f} {3:2.0f} {4:2.0f} {5:7.3f} {6:3.0f} {7:15.6f} {8:2.0f} \n".format(year, mm, dd, hh, minutes, sl, doy, mjd,ss))

    return tt,obstimes,slevel


def pickup_from_noaa(station,date1,date2,datum, printmeta):
    """
    pickup up NOAA data between date1 and date2, which can be 
    longer than one month (NOAA API restriction) 

    Parameters
    ----------
    station: str
        station name

    date1: str 
        beginning time, 20120101 is January 1, 2012

    date2: str
        end time , same format

    datum: str
        what kind of datum is requested

    printmeta : bool
        print metadata to screen 

    Returns
    -------
    data : dictionary in NOAA format

    error : bool

    """

    error = False
    urlL = "https://api.tidesandcurrents.noaa.gov/api/prod/datagetter?"
    # add the datum - i think
    endL = "&product=water_level&datum=" + datum + "&units=metric&time_zone=gmt&application=web_services&format=json"
    url = urlL + "begin_date=" + date1 + "&end_date=" + date2 + "&station=" + station + endL

    try:
        data = requests.get(url).json()
    except: 
        print('There was a major problem with your download.  Maybe internet is off?')
        sys.exit()

    if 'error' in data.keys():
        print(data['error'])
        error = True
    else:
        if printmeta:
            print(data['metadata'])

    return data, error

def download_qld(station,year,plt):
    """
    Parameters
    ----------
    station : str
        tide gauge station name
    year : int
        calendar year
    plt : bool
        whether you want a plot to the screen
    """
    # unfortunately hardwired for 2022 now, so it does not work for other years ...
    #https://www.qld.gov.au/environment/coasts-waterways/beach/storm/storm-sites
    # last seven days
    # https://www.data.qld.gov.au/dataset/coastal-data-system-near-real-time-storm-tide-data
    if year is None:
        print('Year is required for the Queensland option')
        sys.exit()
    else:
        url = 'https://www.data.qld.gov.au/dataset/179c7cc5-26a7-4f57-9e1e-3ec2ed5dd4de/resource/cacc9e98-be38-44f7-a535-07acd22a3b91/'
        url2 = 'download/h071004a_' + str(year) + '_' + station + '_10min.csv'
        tmpfile = station + '_' + str(year) + '_10min.csv'
        wget.download(url+url2, tmpfile)

    xdir = os.environ['REFL_CODE']
    outdir = xdir  + '/Files/'
    if not os.path.exists(outdir) :
        subprocess.call(['mkdir', outdir])
    outfile = outdir + station + '_' + str(year) + '.txt'

    fout = open(outfile, 'w+')
    print('Queensland Tide file written to :', outfile)

    obstimes=[]
    sl = []
    fout.write("{0:s} \n".format('%' + ' QLD Station: ' + station ))
    fout.write("%YYYY MM DD HH MM  SS Water(m) DOY    MJD  \n")
    fout.write("% 1    2  3  4  5   6   7       8     9\n")

    if os.path.isfile(tmpfile):
        x=np.loadtxt(tmpfile,usecols=(0,1,3),skiprows=40,dtype='str',delimiter=',')
        nr,nc=np.shape(x)
        for i in range(0,nr):
            sealevel = float(x[i,2])
            sl.append(sealevel)
            d = int(x[i,0][0:2]) ; m = int(x[i,0][3:5])
            y = int(x[i,0][6:10])
            hr = int(x[i,1][1:3]); mm = int(x[i,1][4:6])
            bigT = datetime.datetime(year=y, month=m, day=d,hour=hr,minute=mm,second=0)
            obstimes.append(bigT)
            modjul, fr = g.mjd(y,m,d,hr,mm,0)
            mjd = modjul+fr
            doy,cdoy,cyyyy,cyy = g.ymd2doy(y, m, d )
            fout.write(" {0:4.0f} {1:2.0f} {2:2.0f} {3:2.0f} {4:2.0f} {5:2.0f} {6:7.3f} {7:3.0f} {8:15.6f} \n".format(y, m, d, hr, mm, 0, sealevel, doy, mjd))

        fout.close()
    else:
        print('No file was downloaded')
        sys.exit()

    if plt:
        g.quickp(station,obstimes,sl)

    # clean up - remove csv file
    subprocess.call(['rm','-f',tmpfile])

# this is more direct API call for last seven days ... I think 
#urlL = 'https://www.data.qld.gov.au/api/3/action/datastore_search?resource_id=7afe7233-fae0-4024-bc98-3a72f05675bd'
#endL = '&q=' + station + '&limit=' + str(NV)

# https://www.data.qld.gov.au/dataset/coastal-data-system-near-real-time-storm-tide-data


if __name__ == "__main__":
    main()

